import trafaret as t

from datarobot.enums import CUSTOM_MODEL_TARGET_TYPE, DEFAULT_MAX_WAIT, NETWORK_EGRESS_POLICY
from datarobot.errors import ClientError
from datarobot.models.api_object import APIObject
from datarobot.models.custom_model_version import CustomModelVersion
from datarobot.utils import encode_utf8_if_py2
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution


class _CustomModelBase(APIObject):
    _path = "customModels/"

    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("name"): t.String(),
            t.Key("description"): t.String(allow_blank=True),
            t.Key("supports_binary_classification", optional=True, default=False): t.Bool(),
            t.Key("supports_regression", optional=True, default=False): t.Bool(),
            t.Key("latest_version", optional=True, default=None): t.Or(
                CustomModelVersion.schema, t.Null()
            ),
            t.Key("deployments_count"): t.Int(),
            t.Key("created_by"): t.String(),
            t.Key("updated") >> "updated_at": t.String(),
            t.Key("created") >> "created_at": t.String(),
            t.Key("target_type", optional=True, default=None): t.String(),
        }
    ).ignore_extra("*")

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return encode_utf8_if_py2(u"{}({!r})".format(self.__class__.__name__, self.name or self.id))

    def _set_values(
        self,
        id,
        name,
        description,
        supports_binary_classification,
        supports_regression,
        latest_version,
        deployments_count,
        created_by,
        updated_at,
        created_at,
        target_type,
    ):
        self.id = id
        self.name = name
        self.description = description

        self.target_type = target_type
        if supports_binary_classification and supports_regression:
            raise ValueError("Model should support only 1 target type")

        if not target_type:
            if supports_binary_classification:
                self.target_type = CUSTOM_MODEL_TARGET_TYPE.BINARY
            elif supports_regression:
                self.target_type = CUSTOM_MODEL_TARGET_TYPE.REGRESSION
            else:
                raise ValueError("Target type must be provided")
        else:
            if target_type != CUSTOM_MODEL_TARGET_TYPE.BINARY and supports_binary_classification:
                raise ValueError(
                    "Cannot specify both target_type {} and "
                    "supports_binary_classification.".format(target_type)
                )
            elif target_type != CUSTOM_MODEL_TARGET_TYPE.REGRESSION and supports_regression:
                raise ValueError(
                    "Cannot specify both target_type {} and "
                    "supports_regression.".format(target_type)
                )

        self.latest_version = CustomModelVersion(**latest_version) if latest_version else None
        self.deployments_count = deployments_count
        self.created_by = created_by
        self.updated_at = updated_at
        self.created_at = created_at

    @classmethod
    def _check_model_type(cls, data):
        return data["customModelType"] == cls._model_type

    @classmethod
    def list(cls, is_deployed=None, order_by=None, search_for=None):
        payload = {
            "custom_model_type": cls._model_type,
            "is_deployed": is_deployed,
            "order_by": order_by,
            "search_for": search_for,
        }
        data = unpaginate(cls._path, payload, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, custom_model_id):
        path = "{}{}/".format(cls._path, custom_model_id)
        data = cls._client.get(path).json()
        if not cls._check_model_type(data):
            raise Exception("Requested model is not a {} model".format(cls._model_type))
        return cls.from_server_data(data)

    def download_latest_version(self, file_path):
        path = "{}{}/download/".format(self._path, self.id)

        response = self._client.get(path)
        with open(file_path, "wb") as f:
            f.write(response.content)

    @classmethod
    def create(cls, name, target_type, description=None, **kwargs):
        payload = dict(
            custom_model_type=cls._model_type, name=name, description=description, **kwargs
        )
        if target_type in CUSTOM_MODEL_TARGET_TYPE.ALL:
            payload["target_type"] = target_type

        # this will be removed when these params are fully deprecated
        if target_type == CUSTOM_MODEL_TARGET_TYPE.BINARY:
            payload["supports_binary_classification"] = True
        elif target_type == CUSTOM_MODEL_TARGET_TYPE.REGRESSION:
            payload["supports_regression"] = True
        elif target_type not in CUSTOM_MODEL_TARGET_TYPE.ALL:
            raise ClientError(
                "Unsupported target_type. target_type must be in  {}.".format(
                    CUSTOM_MODEL_TARGET_TYPE.ALL
                )
            )

        response = cls._client.post(cls._path, data=payload)
        return cls.from_server_data(response.json())

    @classmethod
    def copy_custom_model(
        cls, custom_model_id,
    ):
        path = "{}fromCustomModel/".format(cls._path)
        response = cls._client.post(path, data={"custom_model_id": custom_model_id})
        return cls.from_server_data(response.json())

    def update(self, name=None, description=None, **kwargs):
        payload = dict(name=name, description=description, **kwargs)
        url = "{}{}/".format(self._path, self.id)
        response = self._client.patch(url, data=payload)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))

    def refresh(self):
        url = self._path.format(self.id)
        path = "{}{}/".format(url, self.id)

        response = self._client.get(path)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))

    def delete(self):
        url = "{}{}/".format(self._path, self.id)
        self._client.delete(url)


class CustomInferenceModel(_CustomModelBase):
    """A custom inference model.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        id of the custom model
    name: str
        name of the custom model
    language: str
        programming language of the custom model.
        Can be "python", "r", "java" or "other"
    description: str
        description of the custom model
    target_type: datarobot.TARGET_TYPE
        target type of the custom inference model.
        Values: [`datarobot.TARGET_TYPE.BINARY`, `datarobot.TARGET_TYPE.REGRESSION`,
        `datarobot.TARGET_TYPE.MULTICLASS`, `datarobot.TARGET_TYPE.UNSTRUCTURED`,
        `datarobot.TARGET_TYPE.ANOMALY`]
    target_name: str, optional
        Target feature name;
        it is optional(ignored if provided) for `datarobot.TARGET_TYPE.UNSTRUCTURED`
        or `datarobot.TARGET_TYPE.ANOMALY` target type
    latest_version: datarobot.CustomModelVersion or None
        latest version of the custom model if the model has a latest version
    deployments_count: int
        number of a deployments of the custom models
    target_name: str
        custom model target name
    positive_class_label: str
        for binary classification projects, a label of a positive class
    negative_class_label: str
        for binary classification projects, a label of a negative class
    prediction_threshold: float
        for binary classification projects, a threshold used for predictions
    training_data_assignment_in_progress: bool
        flag describing if training data assignment is in progress
    training_dataset_id: str, optional
        id of a dataset assigned to the custom model
    training_dataset_version_id: str, optional
        id of a dataset version assigned to the custom model
    training_data_file_name: str, optional
        name of assigned training data file
    training_data_partition_column: str, optional
        name of a partition column in a training dataset assigned to the custom model
    created_by: str
        username of a user who user who created the custom model
    updated_at: str
        ISO-8601 formatted timestamp of when the custom model was updated
    created_at: str
        ISO-8601 formatted timestamp of when the custom model was created
    network_egress_policy: datarobot.NETWORK_EGRESS_POLICY, optional
        Determines whether the given custom model is isolated, or can access the public network.
        Can be either 'datarobot.NONE' or 'datarobot.PUBLIC'
    maximum_memory: int, optional
        The maximum memory that might be allocated by the custom-model.
        If exceeded, the custom-model will be killed by k8s
    replicas: int, optional
        A fixed number of replicas that will be deployed in the cluster
    """

    _model_type = "inference"
    _converter = (
        _CustomModelBase._converter
        + {
            t.Key("language"): t.String(allow_blank=True),
            t.Key("target_name", optional=True): t.String(),
            t.Key("training_dataset_id", optional=True): t.String(),
            t.Key("training_dataset_version_id", optional=True): t.String(),
            t.Key("training_data_assignment_in_progress"): t.Bool(),
            t.Key("positive_class_label", optional=True): t.String(),
            t.Key("negative_class_label", optional=True): t.String(),
            t.Key("class_labels", optional=True): t.List(t.String()),
            t.Key("prediction_threshold", optional=True): t.Float(),
            t.Key("training_data_file_name", optional=True): t.String(),
            t.Key("training_data_partition_column", optional=True): t.String(),
            t.Key("network_egress_policy", optional=True): t.Enum(*NETWORK_EGRESS_POLICY.ALL),
            t.Key("maximum_memory", optional=True): t.Int(),
            t.Key("replicas", optional=True): t.Int(),
        }
    ).allow_extra("*")

    def __init__(self, *args, **kwargs):
        super(CustomInferenceModel, self).__init__(*args, **kwargs)

    def __repr__(self):
        return super(CustomInferenceModel, self).__repr__()

    def _set_values(
        self,
        language,
        training_data_assignment_in_progress,
        target_name=None,
        positive_class_label=None,
        negative_class_label=None,
        prediction_threshold=None,
        class_labels=None,
        training_dataset_id=None,
        training_dataset_version_id=None,
        training_data_file_name=None,
        training_data_partition_column=None,
        network_egress_policy=None,
        maximum_memory=None,
        replicas=None,
        **custom_model_kwargs
    ):
        super(CustomInferenceModel, self)._set_values(**custom_model_kwargs)

        self.language = language
        self.target_name = target_name
        self.training_dataset_id = training_dataset_id
        self.training_dataset_version_id = training_dataset_version_id
        self.training_data_assignment_in_progress = training_data_assignment_in_progress
        self.positive_class_label = positive_class_label
        self.negative_class_label = negative_class_label
        self.class_labels = class_labels
        self.prediction_threshold = prediction_threshold
        self.training_data_file_name = training_data_file_name
        self.training_data_partition_column = training_data_partition_column
        self.network_egress_policy = network_egress_policy
        self.maximum_memory = maximum_memory
        self.replicas = replicas

    @classmethod
    def list(
        cls, is_deployed=None, search_for=None, order_by=None,
    ):
        """List custom inference models available to the user.

        .. versionadded:: v2.21

        Parameters
        ----------
        is_deployed: bool, optional
            flag for filtering custom inference models.
            If set to `True`, only deployed custom inference models are returned.
            If set to `False`, only not deployed custom inference models are returned
        search_for: str, optional
            string for filtering custom inference models - only custom
            inference models that contain the string in name or description will
            be returned.
            If not specified, all custom models will be returned
        order_by: str, optional
            property to sort custom inference models by.
            Supported properties are "created" and "updated".
            Prefix the attribute name with a dash to sort in descending order,
            e.g. order_by='-created'.
            By default, the order_by parameter is None which will result in
            custom models being returned in order of creation time descending

        Returns
        -------
        List[CustomInferenceModel]
            a list of custom inference models.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        return super(CustomInferenceModel, cls).list(is_deployed, order_by, search_for)

    @classmethod
    def get(cls, custom_model_id):
        """Get custom inference model by id.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            id of the custom inference model

        Returns
        -------
        CustomInferenceModel
            retrieved custom inference model

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        return super(CustomInferenceModel, cls).get(custom_model_id)

    def download_latest_version(self, file_path):
        """Download the latest custom inference model version.

        .. versionadded:: v2.21

        Parameters
        ----------
        file_path: str
            path to create a file with custom model version content

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        super(CustomInferenceModel, self).download_latest_version(file_path)

    @classmethod
    def create(
        cls,
        name,
        target_type,
        target_name=None,
        language=None,
        description=None,
        positive_class_label=None,
        negative_class_label=None,
        prediction_threshold=None,
        class_labels=None,
        class_labels_file=None,
        network_egress_policy=None,
        maximum_memory=None,
        replicas=None,
    ):
        """Create a custom inference model.

        .. versionadded:: v2.21

        Parameters
        ----------
        name: str
            name of the custom inference model
        target_type: datarobot.TARGET_TYPE
            target type of the custom inference model.
            Values: [`datarobot.TARGET_TYPE.BINARY`, `datarobot.TARGET_TYPE.REGRESSION`,
            `datarobot.TARGET_TYPE.MULTICLASS`, `datarobot.TARGET_TYPE.UNSTRUCTURED`]
        target_name: str, optional
            Target feature name;
            it is optional(ignored if provided) for `datarobot.TARGET_TYPE.UNSTRUCTURED` target type
        language: str, optional
            programming language of the custom learning model
        description: str, optional
            description of the custom learning model
        positive_class_label: str, optional
            custom inference model positive class label for binary classification
        negative_class_label: str, optional
            custom inference model negative class label for binary classification
        prediction_threshold: float, optional
            custom inference model prediction threshold
        class_labels: List[str], optional
            custom inference model class labels for multiclass classification
            Cannot be used with class_labels_file
        class_labels_file: str, optional
            path to file containing newline separated class labels for multiclass classification.
            Cannot be used with class_labels
        network_egress_policy: datarobot.NETWORK_EGRESS_POLICY, optional
            Determines whether the given custom model is isolated, or can access the public network.
            Can be either 'datarobot.NONE' or 'datarobot.PUBLIC'
        maximum_memory: int, optional
            The maximum memory that might be allocated by the custom-model.
            If exceeded, the custom-model will be killed by k8s
        replicas: int, optional
            A fixed number of replicas that will be deployed in the cluster

        Returns
        -------
        CustomInferenceModel
            created a custom inference model

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        if target_type in CUSTOM_MODEL_TARGET_TYPE.REQUIRES_TARGET_NAME and target_name is None:
            raise ValueError(
                "target_name is required for custom models with target type {}".format(target_type)
            )
        if class_labels and class_labels_file:
            raise ValueError("class_labels and class_labels_file cannot be used together")
        if class_labels_file:
            with open(class_labels_file) as f:
                class_labels = [label for label in f.read().split("\n") if label]

        return super(CustomInferenceModel, cls).create(
            name,
            target_type,
            description,
            language=language,
            target_name=target_name,
            positive_class_label=positive_class_label,
            negative_class_label=negative_class_label,
            prediction_threshold=prediction_threshold,
            class_labels=class_labels,
            network_egress_policy=network_egress_policy,
            maximum_memory=maximum_memory,
            replicas=replicas,
        )

    @classmethod
    def copy_custom_model(
        cls, custom_model_id,
    ):
        """Create a custom inference model by copying existing one.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            id of the custom inference model to copy

        Returns
        -------
        CustomInferenceModel
            created a custom inference model

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        return super(CustomInferenceModel, cls).copy_custom_model(custom_model_id)

    def update(
        self,
        name=None,
        language=None,
        description=None,
        target_name=None,
        positive_class_label=None,
        negative_class_label=None,
        prediction_threshold=None,
        class_labels=None,
        class_labels_file=None,
    ):
        """Update custom inference model properties.

        .. versionadded:: v2.21

        Parameters
        ----------
        name: str, optional
            new custom inference model name
        language: str, optional
            new custom inference model programming language
        description: str, optional
            new custom inference model description
        target_name: str, optional
            new custom inference model target name
        positive_class_label: str, optional
            new custom inference model positive class label
        negative_class_label: str, optional
            new custom inference model negative class label
        prediction_threshold: float, optional
            new custom inference model prediction threshold
        class_labels: List[str], optional
            custom inference model class labels for multiclass classification
            Cannot be used with class_labels_file
        class_labels_file: str, optional
            path to file containing newline separated class labels for multiclass classification.
            Cannot be used with class_labels

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        if class_labels and class_labels_file:
            raise ValueError("class_labels and class_labels_file cannot be used together")
        if class_labels_file:
            with open(class_labels_file) as f:
                class_labels = [label for label in f.read().split("\n") if label]

        super(CustomInferenceModel, self).update(
            name,
            description,
            language=language,
            target_name=target_name,
            positive_class_label=positive_class_label,
            negative_class_label=negative_class_label,
            prediction_threshold=prediction_threshold,
            class_labels=class_labels,
        )

    def refresh(self):
        """Update custom inference model with the latest data from server.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        super(CustomInferenceModel, self).refresh()

    def delete(self):
        """Delete custom inference model.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        super(CustomInferenceModel, self).delete()

    def assign_training_data(self, dataset_id, partition_column=None, max_wait=DEFAULT_MAX_WAIT):
        """Assign training data to the custom inference model.

        .. versionadded:: v2.21

        Parameters
        ----------
        dataset_id: str
            the id of the training dataset to be assigned
        partition_column: str, optional
            name of a partition column in the training dataset
        max_wait: int, optional
            max time to wait for a training data assignment.
            If set to None - method will return without waiting.
            Defaults to 10 min

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        payload = {"dataset_id": dataset_id, "partition_column": partition_column}

        path = "{}{}/trainingData/".format(self._path, self.id)

        response = self._client.patch(path, data=payload)

        if max_wait is not None:
            wait_for_async_resolution(self._client, response.headers["Location"], max_wait)

        self.refresh()



import json

import pytest
import responses

import datarobot as dr
from datarobot.utils import camelize
from tests.utils import request_body_to_json


class TestHDFSCreate(object):
    @pytest.fixture
    def hdfs_create_response(self, unittest_endpoint, async_url):
        mysql_create_url = "{}/{}/".format(unittest_endpoint, "hdfsProjects")
        responses.add(
            responses.POST,
            mysql_create_url,
            body="",
            status=202,
            content_type="application/json",
            adding_headers={"Location": async_url},
        )

    @responses.activate
    @pytest.mark.usefixtures("project_creation_responses", "hdfs_create_response")
    def test_create_hdfs_minimum(self):
        dr.Project.create_from_hdfs(url="hdfs://webhdfs.com/directory/file.csv")
        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body["url"] == "hdfs://webhdfs.com/directory/file.csv"

    @responses.activate
    @pytest.mark.usefixtures("project_creation_responses", "hdfs_create_response")
    def test_create_hdfs_all_options(self):
        dr.Project.create_from_hdfs(
            url="hdfs://webhdfs.com/directory/file.csv",
            port=9999,
            project_name="HDFS Project",
            max_wait=100,
        )
        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body["url"] == "hdfs://webhdfs.com/directory/file.csv"
        assert create_body["port"] == 9999
        assert create_body["projectName"] == "HDFS Project"


class TestDataSourceCreate(object):
    @pytest.fixture
    def data_source_create_response(self, unittest_endpoint, async_url):
        project_create_url = "{}/{}".format(unittest_endpoint, dr.Project._path)
        responses.add(
            responses.POST,
            project_create_url,
            body="",
            status=202,
            content_type="application/json",
            adding_headers={"Location": async_url},
        )

    @pytest.fixture
    def data_source(self):
        return dr.DataSource("5acc8437ec8d670001ba16bf")

    @responses.activate
    @pytest.mark.usefixtures("project_creation_responses", "data_source_create_response")
    def test_create_from_data_source_id(self):
        username = "test_user"
        password = "megasuperubersecretpassword"
        data_source_id = "5acc8437ec8d670001ba16bf"
        project_name = "data source project"
        dr.Project.create_from_data_source(
            data_source_id, username, password, project_name=project_name
        )
        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body == {
            "user": username,
            "password": password,
            "dataSourceId": data_source_id,
            "projectName": project_name,
        }


class TestDatasetCreate(object):
    @responses.activate
    @pytest.mark.usefixtures("project_creation_responses")
    def test_create_from_dataset_use_defaults(
        self, project_collection_url, project_without_target_json
    ):
        dataset_id = "abc123"
        project = dr.Project.create_from_dataset(dataset_id)

        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body == {"datasetId": dataset_id}
        assert responses.calls[0].request.url == project_collection_url

        response_json = json.loads(project_without_target_json)
        assert project.id == response_json["id"]
        assert project.project_name == response_json["projectName"]
        assert project.file_name == response_json["fileName"]
        assert project.stage == response_json["stage"]

    @responses.activate
    @pytest.mark.usefixtures("project_creation_responses")
    @pytest.mark.parametrize(
        "param_name,param_value",
        [
            ("dataset_version_id", "20gjkfdlj"),
            ("project_name", "my cool name"),
            ("user", "that guy over there"),
            ("password", "my voice is my passport"),
            ("credential_id", "12345"),
            ("use_kerberos", True),
        ],
    )
    def test_create_from_dataset_using_params(self, param_name, param_value):
        dataset_id = "abc123"
        dr.Project.create_from_dataset(dataset_id, **{param_name: param_value})
        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body == {"datasetId": dataset_id, camelize(param_name): param_value}



# -*- coding: utf-8 -*-
import json
from operator import itemgetter

import pytest
import responses
from trafaret import DataError

from datarobot import errors
from datarobot.models import FeatureAssociationMatrix
from datarobot.utils import from_api


@responses.activate
def test_get(project_id, feature_association_matrix_url, feature_association_matrix):
    responses.add(
        responses.GET,
        feature_association_matrix_url,
        status=200,
        content_type="application/json",
        body=json.dumps(feature_association_matrix),
    )
    feature_association = FeatureAssociationMatrix.get(project_id=project_id)
    assert feature_association
    assert feature_association.project_id == project_id
    assert len(feature_association.strengths) == len(feature_association_matrix["strengths"])
    assert len(feature_association.features) == len(feature_association_matrix["features"])
    assert sorted(feature_association.strengths, key=itemgetter("feature1")) == sorted(
        feature_association_matrix["strengths"], key=itemgetter("feature1")
    )
    assert sorted(feature_association.features, key=itemgetter("feature")) == sorted(
        from_api(feature_association_matrix["features"]), key=itemgetter("feature")
    )


@responses.activate
def test_get_for_featurelist(
    project_id, feature_association_matrix_url, feature_association_matrix
):
    featurelist_id = "6033e2b9ecb9108111e4c677"
    url = feature_association_matrix_url + "?featurelistId={}".format(featurelist_id)
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(feature_association_matrix),
    )
    feature_association = FeatureAssociationMatrix.get(
        project_id=project_id, featurelist_id=featurelist_id
    )
    assert feature_association
    assert feature_association.project_id == project_id
    assert len(feature_association.strengths) == len(feature_association_matrix["strengths"])
    assert len(feature_association.features) == len(feature_association_matrix["features"])
    assert sorted(feature_association.strengths, key=itemgetter("feature1")) == sorted(
        feature_association_matrix["strengths"], key=itemgetter("feature1")
    )
    assert sorted(feature_association.features, key=itemgetter("feature")) == sorted(
        from_api(feature_association_matrix["features"]), key=itemgetter("feature")
    )


@responses.activate
def test_get__not_found(project_id, feature_association_matrix_url):
    responses.add(responses.GET, feature_association_matrix_url, status=404)
    with pytest.raises(errors.ClientError, match="404 client error: "):
        FeatureAssociationMatrix.get(project_id=project_id)


@pytest.mark.parametrize("project_id", [None, "", 1, 42.1])
def test_get__invalid_project_id(project_id):
    with pytest.raises(DataError):
        FeatureAssociationMatrix.get(project_id)


def test_get__invalid_arguments(project_id):
    with pytest.raises(DataError):
        FeatureAssociationMatrix.get(
            project_id, metric="fake", association_type="fake", featurelist_id="",
        )



import json

from mock import patch
import pytest
import responses

from datarobot import AppPlatformError, errors, Project
from datarobot.enums import DEFAULT_TIMEOUT
from tests.utils import assert_raised_regex, SDKTestcase


class Test400LevelErrors(SDKTestcase):
    @responses.activate
    def test_401_error_message(self):
        """For whatever reason, the `requests` library completely ignores the
        body in a 401 error, so, here is our own message
        """
        responses.add(
            responses.GET, "https://host_name.com/projects/", status=401,
        )

        with pytest.raises(AppPlatformError) as exc_info:
            Project.list()
        assert_raised_regex(exc_info, "not properly authenticated")

    @responses.activate
    def test_404_error_message(self):
        """Check that we handle correctly 4xx error with empty string in response body
        """
        responses.add(responses.GET, "https://host_name.com/projects/", status=404, body="")

        with pytest.raises(AppPlatformError) as exc_info:
            Project.list()
        expected_error = "{} client error: {}".format(404, "")
        assert str(exc_info.value) == expected_error

    @responses.activate
    def test_403_error_message(self):
        """ Check that we output same error message that server returns on 403
        """
        error_text = "Some permission error"
        error_response = {"message": error_text}

        responses.add(
            responses.GET,
            "https://host_name.com/projects/",
            status=403,
            body=json.dumps(error_response),
        )

        with pytest.raises(AppPlatformError) as exc_info:
            Project.list()
        expected_error = "{} client error: {}".format(403, json.dumps(error_response))
        assert str(exc_info.value) == expected_error

    @responses.activate
    def test_model_already_added_exception(self):
        responses.add(
            responses.POST,
            "https://host_name.com/projects/p-id/models/",
            status=422,
            body=json.dumps({"message": u"Model already added", "errorName": u"JobAlreadyAdded"}),
        )

        with pytest.raises(errors.JobAlreadyRequested):
            Project("p-id").train("some-blueprint-id")

    @responses.activate
    def test_extracts_error_message(self):
        data = {"message": "project p-id has been deleted"}
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/models/",
            status=422,
            body=json.dumps(data),
            content_type="application/json",
        )

        with pytest.raises(AppPlatformError) as exc_info:
            Project("p-id").get_models()
        assert_raised_regex(exc_info, "project p-id has been deleted")


def test_client_uses_both_timeouts(client):
    connect_timeout = 33
    client.connect_timeout = connect_timeout

    with patch("datarobot.rest.requests.Session.request") as request_mock:
        client.get("ping")
        args, kwargs = request_mock.call_args
        called_timeout = kwargs["timeout"]
        assert called_timeout == (connect_timeout, DEFAULT_TIMEOUT.READ)


@responses.activate
def test_client_passes_read_timeout_from_build_request(client, temporary_file):
    connect_timeout = 33
    read_timeout = 180
    client.connect_timeout = connect_timeout

    with patch("datarobot.rest.requests.Session.request") as request_mock:
        client.build_request_with_file(
            "POST",
            "file_upload",
            "my_file.csv",
            file_path=temporary_file,
            read_timeout=read_timeout,
        )
        args, kwargs = request_mock.call_args
        called_timeout = kwargs["timeout"]
        assert called_timeout == (connect_timeout, read_timeout)



#! /usr/bin/env python3
"""Show a small sample of images and associated activation maps images.

The following will open a project, get the first model id where the feature
name matches, and then get a list of the activation maps. Then it will
display a few of the images and the associated images with overlay in the
GUI.

The parameters may be adjusted to use your project name, feature name, and
the number of images to display.
"""
import io

import PIL.Image

from datarobot.models import Project
from datarobot.models.visualai import ImageActivationMap


def display_images(project_name, feature_name, max_images):
    project = Project.list(search_params={"project_name": project_name})[0]
    model_id = next(
        mid for mid, name in ImageActivationMap.models(project.id) if name == feature_name
    )
    for amap in ImageActivationMap.list(project.id, model_id, feature_name)[:max_images]:
        with io.BytesIO(amap.image.image_bytes) as bio, PIL.Image.open(bio) as img:
            img.show()
        with io.BytesIO(amap.overlay_image.image_bytes) as bio, PIL.Image.open(bio) as img:
            img.show()


if __name__ == "__main__":
    project_name = "dataset_2k.zip"
    feature_name = "image"
    max_images = 2
    display_images(project_name, feature_name, max_images)



from copy import deepcopy
import json

import pytest
import responses

from datarobot import CustomInferenceImage
from datarobot.utils import camelize


@pytest.fixture
def mocked_test():
    return {
        "id": "5cf4d3f5f930e26daac18alt",
        "status": "succeeded",
        "completedAt": "2019-09-28T15:19:26.587583Z",
    }


@pytest.fixture
def mocked_images(mocked_test):
    return {
        "count": 2,
        "totalCount": 2,
        "next": None,
        "previous": None,
        "data": [
            {
                "id": "5cf4d3f5f930e26daac18a1a",
                "customModel": {"id": "5cf4d3f5f930e26daac18acc", "name": "cm"},
                "customModelVersion": {"id": "5cf4d3f5f930e26daac18acv", "label": "cmv"},
                "executionEnvironment": {"id": "5cf4d3f5f930e26daac18aee", "name": "ee"},
                "executionEnvironmentVersion": {"id": "5cf4d3f5f930e26daac18aev", "label": "eev"},
                "latestTest": mocked_test,
            },
            {
                "id": "5cf4d3f5f930e26daac18a2b",
                "customModel": {"id": "5cf4d3f5f930e26daac18bcc", "name": "cm n2"},
                "customModelVersion": {"id": "5cf4d3f5f930e26daac18bcv", "label": "cmv n2"},
                "executionEnvironment": {"id": "5cf4d3f5f930e26daac18bee", "name": "ee n2"},
                "executionEnvironmentVersion": {
                    "id": "5cf4d3f5f930e26daac18bev",
                    "label": "eev n2",
                },
                "latestTest": None,
            },
        ],
    }


@pytest.fixture
def mocked_image(mocked_images):
    return mocked_images["data"][0]


@pytest.fixture
def make_images_url(unittest_endpoint):
    def _make_images_url(image_id=None):
        base_url = "{}/customInferenceImages/".format(unittest_endpoint)
        if image_id is not None:
            return "{}{}/".format(base_url, image_id)
        return base_url

    return _make_images_url


def assert_image(image, image_json):
    assert image.id == image_json["id"]
    assert image.custom_model["id"] == image_json["customModel"]["id"]
    assert image.custom_model["name"] == image_json["customModel"]["name"]
    assert image.custom_model_version["id"] == image_json["customModelVersion"]["id"]
    assert image.custom_model_version["label"] == image_json["customModelVersion"]["label"]
    assert image.execution_environment["id"] == image_json["executionEnvironment"]["id"]
    assert image.execution_environment["name"] == image_json["executionEnvironment"]["name"]

    exec_env_ver = image_json["executionEnvironmentVersion"]
    assert image.execution_environment_version["id"] == exec_env_ver["id"]
    assert image.execution_environment_version["label"] == exec_env_ver["label"]
    if image.latest_test is None:
        assert image_json["latestTest"] is None
    else:
        assert image.latest_test["id"] == image_json["latestTest"]["id"]
        assert image.latest_test["status"] == image_json["latestTest"]["status"]
        assert image.latest_test["completed_at"] == image_json["latestTest"]["completedAt"]


def mock_get_response(url, response):
    responses.add(
        responses.GET, url, status=200, content_type="application/json", body=json.dumps(response),
    )


def test_from_server_data(mocked_image):
    image = CustomInferenceImage.from_server_data(mocked_image)
    assert_image(image, mocked_image)


@responses.activate
def test_create_image(mocked_image, make_images_url, tmpdir):
    responses.add(
        responses.POST,
        make_images_url(),
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_image),
    )

    image = CustomInferenceImage.create(
        mocked_image["customModel"]["id"],
        mocked_image["customModelVersion"]["id"],
        mocked_image["executionEnvironment"]["id"],
        mocked_image["executionEnvironmentVersion"]["id"],
    )
    assert_image(image, mocked_image)

    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == make_images_url()


@responses.activate
def test_get_image(mocked_image, make_images_url):
    url = make_images_url(mocked_image["id"])
    mock_get_response(url, mocked_image)

    image = CustomInferenceImage.get(mocked_image["id"])
    assert_image(image, mocked_image)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_images(mocked_images, make_images_url):
    url = make_images_url()
    mock_get_response(url, mocked_images)

    images = CustomInferenceImage.list()

    assert len(images) == len(mocked_images["data"])
    for image, mocked_image in zip(images, mocked_images["data"]):
        assert_image(image, mocked_image)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_images_multiple_pages(mocked_images, make_images_url):
    url1 = make_images_url()
    url2 = make_images_url() + "2"

    mocked_images_2nd = deepcopy(mocked_images)
    mocked_images["next"] = url2

    mock_get_response(url1, mocked_images)
    mock_get_response(url2, mocked_images_2nd)

    images = CustomInferenceImage.list()
    assert len(images) == len(mocked_images["data"]) + len(mocked_images_2nd["data"])
    for image, mocked_image in zip(images, mocked_images["data"] + mocked_images_2nd["data"]):
        assert_image(image, mocked_image)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url.endswith(url1)
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url.endswith(url2)


@responses.activate
@pytest.mark.parametrize(
    "filter, property_key",
    [
        ("testing_status", "latestTest.status"),
        ("custom_model_id", "customModel.id"),
        ("custom_model_version_id", "customModelVersion.id"),
        ("environment_id", "executionEnvironment.id"),
        ("environment_version_id", "executionEnvironmentVersion.id"),
    ],
)
def test_list_filter_build_status(mocked_image, make_images_url, filter, property_key):
    prop = mocked_image
    for key in property_key.split("."):
        prop = prop[key]

    url = make_images_url() + "?{}={}".format(camelize(filter), prop)
    mock_get_response(url, {"count": 1, "next": None, "previous": None, "data": [mocked_image]})

    images = CustomInferenceImage.list(**{filter: prop})

    assert len(images) == 1
    assert_image(images[0], mocked_image)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_refresh(mocked_image, make_images_url, mocked_test):
    # arrange
    url = make_images_url(mocked_image["id"])

    mocked_image.update({"latestTest": None})
    mock_get_response(url, mocked_image)

    mocked_image.update({"latestTest": mocked_test})
    mock_get_response(url, mocked_image)

    # act
    image = CustomInferenceImage.get(mocked_image["id"])
    image.refresh()

    # assert
    assert_image(image, mocked_image)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url


@responses.activate
def test_get_feature_impact(mocked_image, make_images_url, feature_impact_server_data):
    # arrange
    url = make_images_url(mocked_image["id"])
    mock_get_response(url, mocked_image)
    responses.add(
        responses.GET,
        url + "featureImpact/",
        status=200,
        content_type="application/json",
        body=json.dumps(feature_impact_server_data),
    )

    # act
    image = CustomInferenceImage.get(mocked_image["id"])
    feature_impacts = image.get_feature_impact()

    # assert
    assert feature_impacts == feature_impact_server_data["featureImpacts"]

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url + "featureImpact/"


@responses.activate
def test_get_feature_impact_with_metadata(
    mocked_image, make_images_url, feature_impact_server_data, feature_impact_server_data_filtered
):
    # arrange
    url = make_images_url(mocked_image["id"])
    mock_get_response(url, mocked_image)
    responses.add(
        responses.GET,
        url + "featureImpact/",
        status=200,
        content_type="application/json",
        body=json.dumps(feature_impact_server_data),
    )

    # act
    image = CustomInferenceImage.get(mocked_image["id"])
    feature_impacts = image.get_feature_impact(with_metadata=True)

    # assert
    assert feature_impacts == feature_impact_server_data_filtered

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url + "featureImpact/"


@responses.activate
def test_calculate_feature_impact(unittest_endpoint, mocked_image, make_images_url):
    # arrange
    status_url = "{}/status_url".format(unittest_endpoint)
    url = make_images_url(mocked_image["id"])
    impact_url = url + "featureImpact/"
    mock_get_response(url, mocked_image)

    responses.add(
        responses.POST,
        impact_url,
        status=200,
        content_type="application/json",
        headers={"Location": status_url},
        # this ID is ignored, we only use headers and we dont' have a real ID from the backend.
        body=json.dumps({"status_id": "5cf4d3f5f930e26daac18a1a"}),
    )

    responses.add(
        responses.GET, status_url, headers={"Location": url}, status=303,
    )

    # act
    image = CustomInferenceImage.get(mocked_image["id"])
    image.calculate_feature_impact()

    # assert

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "POST"
    assert responses.calls[1].request.url == impact_url
    assert responses.calls[2].request.method == "GET"
    assert responses.calls[2].request.url == status_url



import json
from tempfile import NamedTemporaryFile

import pytest
import responses

from datarobot import Deployment
from datarobot.utils import from_api


class TestGetPredictionResults(object):
    @pytest.fixture
    def response_data(self, deployment_data):
        return {
            "count": 40,
            "next": None,
            "previous": None,
            "associationIdColumnNames": ["col_a"],
            "data": [
                {
                    "timestamp": "2021-03-22 20:00:00+00:00",
                    "associationId": "d6050aa2-2bd3-4ac5-b235-6a1eaa9f11b4",
                    "actual": 0.12,
                    "predicted": 0.15,
                    "modelId": "6059008f904afffa1379d2a6",
                },
                {
                    "timestamp": "2021-03-22 20:00:00+00:00",
                    "associationId": "d6050aa2-2bd3-4ac5-b235-6a1eaa9f11b4",
                    "actual": None,
                    "predicted": "False",
                    "modelId": "6059008f904afffa1379d2a6",
                },
            ],
        }

    @pytest.fixture
    def response(self, unittest_endpoint, deployment_data, response_data):
        url = "{}/deployments/{}/predictionResults/".format(
            unittest_endpoint, deployment_data["id"]
        )
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(response_data),
        )

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "response")
    def test_retrieve(self, deployment_data, response_data):
        actual_results = Deployment.get(deployment_data["id"]).get_prediction_results()
        expected_results = from_api(response_data["data"], keep_null_keys=True)
        assert actual_results == expected_results


class TestDownloadPredictionResults(object):
    @pytest.fixture
    def response_data(self, deployment_data):
        return b"some_data"

    @pytest.fixture
    def response(self, unittest_endpoint, deployment_data, response_data):
        url = "{}/deployments/{}/predictionResults/".format(
            unittest_endpoint, deployment_data["id"]
        )
        responses.add(
            responses.GET, url, status=200, content_type="text/csv", body=response_data,
        )

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "response")
    def test_download(self, deployment_data, response_data):
        with NamedTemporaryFile() as file:
            Deployment.get(deployment_data["id"]).download_prediction_results(file.name)
            assert file.read() == response_data



import json

import mock
import pytest
import responses

from datarobot import CustomModelVersionDependencyBuild
from tests.model.utils import assert_custom_model_version_dependency_build


def mock_response(url, method, response, content_type="application/json"):
    if content_type == "application/json":
        body = json.dumps(response)
    else:
        body = response
    responses.add(
        method, url, status=200, content_type=content_type, body=body,
    )


@pytest.fixture
def make_dependency_build_url(unittest_endpoint):
    def _make_versions_url(custom_model_id, custom_model_version_id, is_log_route=False):
        base_url = "{}/customModels/{}/versions/{}/dependencyBuild{}/".format(
            unittest_endpoint,
            custom_model_id,
            custom_model_version_id,
            "Log" if is_log_route else "",
        )
        return base_url

    return _make_versions_url


@pytest.fixture
def custom_model_ids():
    return "5f2095d1e9790c51a09e6f69", "5f2095d1e9790c51a09e6f69"


@responses.activate
@pytest.mark.parametrize(
    "data",
    [
        "mock_custom_model_dependency_build_submitted",
        "mock_custom_model_dependency_build_processing",
        "mock_custom_model_dependency_build_failed",
        "mock_custom_model_dependency_build_success",
    ],
)
def test_get_info(request, data, custom_model_ids, make_dependency_build_url):
    # arrange
    model_id, version_id = custom_model_ids
    data = request.getfixturevalue(data)
    url = make_dependency_build_url(model_id, version_id)
    mock_response(url, responses.GET, data)

    # act
    build_info = CustomModelVersionDependencyBuild.get_build_info(model_id, version_id)

    # assert
    assert_custom_model_version_dependency_build(build_info, data, model_id, version_id)
    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_build_non_blocking(
    mock_custom_model_dependency_build_submitted, custom_model_ids, make_dependency_build_url
):
    # arrange
    model_id, version_id = custom_model_ids
    url = make_dependency_build_url(model_id, version_id)
    mock_response(url, responses.POST, mock_custom_model_dependency_build_submitted)

    # act
    build_info = CustomModelVersionDependencyBuild.start_build(model_id, version_id, max_wait=None)

    # assert
    assert_custom_model_version_dependency_build(
        build_info, mock_custom_model_dependency_build_submitted, model_id, version_id
    )
    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == url


@responses.activate
@pytest.mark.parametrize(
    "final_data",
    ["mock_custom_model_dependency_build_success", "mock_custom_model_dependency_build_failed"],
)
def test_build_blocking(
    request,
    final_data,
    mock_custom_model_dependency_build_submitted,
    mock_custom_model_dependency_build_processing,
    custom_model_ids,
    make_dependency_build_url,
):
    # arrange
    model_id, version_id = custom_model_ids
    final_data = request.getfixturevalue(final_data)
    url = make_dependency_build_url(model_id, version_id)
    mock_response(url, responses.POST, mock_custom_model_dependency_build_submitted)
    mock_response(url, responses.GET, mock_custom_model_dependency_build_processing)
    mock_response(url, responses.GET, final_data)

    # act
    with mock.patch("time.sleep"):
        build_info = CustomModelVersionDependencyBuild.start_build(model_id, version_id)

    # assert
    assert_custom_model_version_dependency_build(build_info, final_data, model_id, version_id)
    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url
    assert responses.calls[2].request.method == "GET"
    assert responses.calls[2].request.url == url


@responses.activate
def test_get_log(
    mock_custom_model_dependency_build_success, custom_model_ids, make_dependency_build_url
):
    # arrange
    model_id, version_id = custom_model_ids
    url = make_dependency_build_url(model_id, version_id)
    log_url = make_dependency_build_url(model_id, version_id, is_log_route=True)
    expected_log = "this is fine. everything is fine."
    mock_response(url, responses.GET, mock_custom_model_dependency_build_success)
    mock_response(log_url, responses.GET, expected_log, content_type="text/plain")

    # act
    build_info = CustomModelVersionDependencyBuild.get_build_info(model_id, version_id)
    log = build_info.get_log()

    # assert
    assert log == expected_log
    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == log_url


@responses.activate
def test_cancel(
    mock_custom_model_dependency_build_success, custom_model_ids, make_dependency_build_url
):
    # arrange
    model_id, version_id = custom_model_ids
    url = make_dependency_build_url(model_id, version_id)
    mock_response(url, responses.GET, mock_custom_model_dependency_build_success)
    mock_response(url, responses.DELETE, None)

    # act
    build_info = CustomModelVersionDependencyBuild.get_build_info(model_id, version_id)
    build_info.cancel()

    # assert
    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "DELETE"
    assert responses.calls[1].request.url == url


@responses.activate
def test_refresh(
    mock_custom_model_dependency_build_processing,
    mock_custom_model_dependency_build_success,
    custom_model_ids,
    make_dependency_build_url,
):
    # arrange
    model_id, version_id = custom_model_ids
    url = make_dependency_build_url(model_id, version_id)
    mock_response(url, responses.GET, mock_custom_model_dependency_build_processing)
    mock_response(url, responses.GET, mock_custom_model_dependency_build_success)

    # act
    build_info = CustomModelVersionDependencyBuild.get_build_info(model_id, version_id)
    build_info.refresh()

    # assert
    assert_custom_model_version_dependency_build(
        build_info, mock_custom_model_dependency_build_success, model_id, version_id
    )

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url



from mock import patch
import pytest
import responses

from datarobot import ExternalRocCurve
from datarobot.errors import ClientError
from datarobot.utils import from_api


class TestExternalRocCurve(object):
    @staticmethod
    def create_response_data(roc_curve_data, count):
        return [dict(roc_curve_data, datasetId="dataset_id_{}".format(i)) for i in range(count)]

    @pytest.fixture
    def url(self, project_id, model_id):
        return "https://host_name.com/projects/{}/models/{}/datasetRocCurves/".format(
            project_id, model_id
        )

    def test_instantiation(self, roc_curve_data):
        one_roc_chart = self.create_response_data(roc_curve_data, 1)[0]
        roc = ExternalRocCurve.from_server_data(one_roc_chart)

        assert roc.dataset_id == one_roc_chart["datasetId"]
        assert roc.negative_class_predictions == roc_curve_data["negativeClassPredictions"]
        assert roc.positive_class_predictions == roc_curve_data["positiveClassPredictions"]
        assert roc.roc_points == from_api(roc_curve_data["rocPoints"])

    def test_future_proof(self, roc_curve_data):
        one_roc_chart = self.create_response_data(roc_curve_data, 1)[0]
        data_with_future_keys = dict(one_roc_chart, new_key="some future lift data")
        data_with_future_keys["rocPoints"][0]["new_key"] = "some future bin data"
        ExternalRocCurve.from_server_data(data_with_future_keys)

    @responses.activate
    @patch(
        "datarobot.models.external_dataset_scores_insights.external_roc_curve."
        "DEFAULT_BATCH_SIZE",
        10,
    )
    def test_list_roc_all_results(self, roc_curve_data, url, project_id, model_id):
        url_page1 = url + "?limit=10&offset=0"
        url_page2 = url + "?limit=10&offset=10"
        all_data = self.create_response_data(roc_curve_data, 11)
        resp_data_1 = dict(count=10, previous=None, next=url_page2, data=all_data[:10])
        resp_data_2 = dict(count=1, previous=url_page1, next=None, data=all_data[10:])
        responses.add(
            responses.GET, url_page1, json=resp_data_1, status=200, match_querystring=True
        )
        responses.add(
            responses.GET, url_page2, json=resp_data_2, status=200, match_querystring=True
        )

        roc_charts = ExternalRocCurve.list(project_id, model_id, limit=0)
        assert len(roc_charts) == 11
        for i, chart in enumerate(roc_charts):
            assert isinstance(chart, ExternalRocCurve)
            assert chart.dataset_id == "dataset_id_{}".format(i)
            assert chart.negative_class_predictions == roc_curve_data["negativeClassPredictions"]
            assert chart.positive_class_predictions == roc_curve_data["positiveClassPredictions"]
            assert chart.roc_points == from_api(roc_curve_data["rocPoints"])

    @responses.activate
    @patch(
        "datarobot.models.external_dataset_scores_insights.external_roc_curve."
        "DEFAULT_BATCH_SIZE",
        10,
    )
    def test_roc_charts_all_results_with_offset(self, roc_curve_data, url, project_id, model_id):
        url_page1 = url + "?limit=10&offset=3"
        url_page2 = url + "?limit=10&offset=13"
        all_data = self.create_response_data(roc_curve_data, 15)
        resp_data_1 = dict(count=10, previous=None, next=url_page2, data=all_data[3:13])
        resp_data_2 = dict(count=1, previous=url_page1, next=None, data=all_data[13:])
        responses.add(
            responses.GET, url_page1, json=resp_data_1, status=200, match_querystring=True
        )
        responses.add(
            responses.GET, url_page2, json=resp_data_2, status=200, match_querystring=True
        )

        roc_charts = ExternalRocCurve.list(project_id, model_id, limit=0, offset=3)
        assert len(roc_charts) == 12
        assert roc_charts[0].dataset_id == "dataset_id_3"

    @responses.activate
    def test_list_roc_chart_limited(self, roc_curve_data, url, project_id, model_id):
        url_page = url + "?limit=3&offset=2"
        all_data = self.create_response_data(roc_curve_data, 5)
        resp_data = dict(count=3, previous=None, next=None, data=all_data[2:5])
        responses.add(responses.GET, url_page, json=resp_data, status=200, match_querystring=True)
        roc_charts = ExternalRocCurve.list(project_id, model_id, limit=3, offset=2)
        assert len(roc_charts) == 3
        for i, chart in enumerate(roc_charts):
            assert chart.dataset_id == "dataset_id_{}".format(i + 2)
            assert chart.negative_class_predictions == roc_curve_data["negativeClassPredictions"]
            assert chart.positive_class_predictions == roc_curve_data["positiveClassPredictions"]
            assert chart.roc_points == from_api(roc_curve_data["rocPoints"])

    @responses.activate
    def test_get_roc_chart_success(self, roc_curve_data, url, project_id, model_id):
        dataset_id = "dataset_id_0"
        url_page = url + "?datasetId={}&limit=100&offset=0".format(dataset_id)
        all_data = self.create_response_data(roc_curve_data, 2)
        resp_data = dict(count=1, previous=None, next=None, data=[all_data[0]])
        responses.add(responses.GET, url_page, json=resp_data, status=200, match_querystring=True)
        chart = ExternalRocCurve.get(project_id, model_id, dataset_id)
        assert isinstance(chart, ExternalRocCurve)
        assert chart.dataset_id == "dataset_id_0"
        assert chart.negative_class_predictions == roc_curve_data["negativeClassPredictions"]
        assert chart.positive_class_predictions == roc_curve_data["positiveClassPredictions"]
        assert chart.roc_points == from_api(roc_curve_data["rocPoints"])

    @responses.activate
    def test_get_roc_chart_failure(self, url, project_id, model_id):
        dataset_id = "dataset_id_0"
        url_page = url + "?datasetId={}&limit=100&offset=0".format(dataset_id)
        resp_data = dict(count=0, previous=None, next=None, data=[])
        responses.add(responses.GET, url_page, json=resp_data, status=200, match_querystring=True)
        with pytest.raises(ClientError, match="Requested roc curve does not exist."):
            ExternalRocCurve.get(project_id, model_id, dataset_id)

    def test_get_best_f1_threshold(self, roc_curve_data):
        one_roc_chart = self.create_response_data(roc_curve_data, 1)[0]
        roc = ExternalRocCurve.from_server_data(one_roc_chart)
        best_threshold = roc.get_best_f1_threshold()
        best_f1 = roc.estimate_threshold(best_threshold)["f1_score"]
        assert all(best_f1 >= roc_point["f1_score"] for roc_point in roc.roc_points)

    def test_estimate_threshold_equal(self, roc_curve_data):
        one_roc_chart = self.create_response_data(roc_curve_data, 1)[0]
        roc = ExternalRocCurve.from_server_data(one_roc_chart)
        threshold = roc_curve_data["rocPoints"][1]["threshold"]
        assert roc.estimate_threshold(threshold)["threshold"] == threshold

        threshold = roc_curve_data["rocPoints"][1]["threshold"] + 0.1
        assert roc.estimate_threshold(threshold)["threshold"] > threshold



import json

import pytest
import responses

from datarobot import ImportedModel
from tests.test_helpers import fixture_file_path

imported_model_dict = {
    u"originUrl": u"http://app.datarobot.com/",
    u"datasetName": u"train_data.csv",
    u"modelName": u"Linear Regression (5)",
    u"displayName": u"Linear Regression (5)",
    u"target": u"target",
    u"projectId": u"57b37876c808912b1c87094a",
    u"importedByUsername": u"david.chudzicki@datarobot.com",
    u"note": None,
    u"importedAt": u"2016-08-28 22:39:56.117000",
    u"version": 1.8,
    u"projectName": u"Unnamed Project",
    u"featurelistName": u"week_day",
    u"createdByUsername": u"david.chudzicki@datarobot.com",
    u"modelId": u"57b70cf5c80891295500c6e8",
    u"importedById": u"56a5a34b16f4b622fb8426f6",
    u"id": u"301d0dd18e034c058dbed3bdf8207a4a",
    u"createdById": u"56a5a34b16f4b622fb8426f6",
}


def add_imported_model_get_response(model_id):
    url = "https://host_name.com/importedModels/{}/".format(model_id)
    responses.add(responses.GET, url, body=json.dumps(imported_model_dict))


@pytest.fixture
@responses.activate
def imported_model():
    add_imported_model_get_response(imported_model_dict["id"])
    return ImportedModel.get(imported_model_dict["id"])


@responses.activate
def test_get(imported_model):
    assert imported_model.id == imported_model_dict["id"]


@responses.activate
def test_list():
    response_dict = {u"count": 1, u"next": None, u"previous": None, u"data": [imported_model_dict]}
    url = "https://host_name.com/importedModels/"
    responses.add(responses.GET, url, body=json.dumps(response_dict))
    imported_models = ImportedModel.list(imported_model_dict["id"])
    assert isinstance(imported_models, list)
    assert len(imported_models) == 1
    assert isinstance(imported_models[0], ImportedModel)


@responses.activate
def test_update(imported_model):
    url = "https://host_name.com/importedModels/{}/".format(imported_model_dict["id"])
    responses.add(responses.PATCH, url, body="this is not actually used")
    new_note = "new note"
    new_display_name = "new_display_name"
    imported_model.update(display_name=new_display_name, note=new_note)
    assert imported_model.note == new_note
    assert imported_model.display_name == new_display_name


@responses.activate
def test_delete(imported_model):
    url = "https://host_name.com/importedModels/{}/".format(imported_model_dict["id"])
    responses.add(responses.DELETE, url, body="not used")
    imported_model.delete()


@responses.activate
def test_create():
    responses.add(
        responses.POST,
        "https://host_name.com/importedModels/",
        body="",
        status=202,
        content_type="application/json",
        adding_headers={"Location": "https://host_name.com/status/status-id/"},
    )

    responses.add(
        responses.GET,
        "https://host_name.com/status/status-id/",
        status=303,
        body="",
        content_type="application/json",
        adding_headers={
            "Location": "https://host_name.com/importedModels/{}/".format(imported_model_dict["id"])
        },
    )
    add_imported_model_get_response(imported_model_dict["id"])

    imported_model = ImportedModel.create(fixture_file_path("some_model.drmodel"))
    assert isinstance(imported_model, ImportedModel)



import pytest

from datarobot import Job


@pytest.mark.parametrize("is_blocked", [False, True])
def test_job_initialization_with_is_blocked(is_blocked):
    job = Job(
        {
            "status": "queue",
            "url": "https://host_name.com/projects/p-id/modelJobs/1/",
            "id": "1",
            "jobType": "model",
            "projectId": "p-id",
            "isBlocked": is_blocked,
        }
    )

    assert job.id == 1
    assert job.status == "queue"
    assert job.job_type == "model"
    assert job.project_id == "p-id"
    assert job.is_blocked is is_blocked



# flake8: noqa
# because the unused imports are on purpose

from .models.custom_training_blueprint import CustomTrainingBlueprint
from .models.custom_training_model import CustomTrainingModel
from .models.user_blueprints.models import UserBlueprint



from datetime import datetime
import json

from dateutil.tz import tzutc
import pytest
import responses

from datarobot import DatetimeModel
from tests.utils import request_body_to_json


@pytest.fixture
def accuracy_over_time_plots_metadata_response():
    return {
        "holdoutStatuses": {"training": "notCompleted", "validation": "completed"},
        "holdoutMetadata": {
            "training": {"startDate": None, "endDate": None},
            "validation": {
                "startDate": "2014-04-09T00:00:00.000000Z",
                "endDate": "2014-06-15T00:00:00.000000Z",
            },
        },
        "backtestMetadata": [
            {
                "training": {"startDate": None, "endDate": None},
                "validation": {"startDate": None, "endDate": None},
            },
            {
                "training": {"startDate": None, "endDate": None},
                "validation": {
                    "startDate": "2013-11-26T00:00:00.000000Z",
                    "endDate": "2014-02-01T00:00:00.000000Z",
                },
            },
        ],
        "forecastDistance": 1,
        "backtestStatuses": [
            {"training": "errored", "validation": "insufficientData"},
            {"training": "notCompleted", "validation": "completed"},
        ],
        "resolutions": ["days", "weeks", "months", "quarters", "years"],
    }


@pytest.fixture
def accuracy_over_time_plot_response():
    return {
        "startDate": "2014-02-01T00:00:00.000000Z",
        "statistics": {"durbinWatson": 0.5},
        "endDate": "2014-02-03T00:00:00.000000Z",
        "resolution": "days",
        "bins": [
            {
                "startDate": "2014-02-01T00:00:00.000000Z",
                "frequency": 10.0,
                "actual": 1.11,
                "predicted": 1.12,
                "endDate": "2014-02-02T00:00:00.000000Z",
            },
            {
                "startDate": "2014-02-02T00:00:00.000000Z",
                "frequency": 0.0,
                "actual": None,
                "predicted": None,
                "endDate": "2014-02-03T00:00:00.000000Z",
            },
        ],
        "calendarEvents": [
            {
                "date": "2014-02-17T00:00:00.000000Z",
                "seriesId": None,
                "name": "Washington's Birthday",
            }
        ],
    }


@pytest.fixture
def datetime_trend_plot_preview_response():
    return {
        "startDate": "2014-02-01T00:00:00.000000Z",
        "endDate": "2014-04-09T00:00:00.000000Z",
        "bins": [
            {
                "actual": 100,
                "predicted": 101,
                "startDate": "2014-02-01T00:00:00.000000Z",
                "endDate": "2014-02-02T00:00:00.000000Z",
            },
            {
                "actual": 100,
                "predicted": 101,
                "startDate": "2014-04-08T00:00:00.000000Z",
                "endDate": "2014-04-09T00:00:00.000000Z",
            },
        ],
    }


@pytest.fixture
def forecast_vs_actual_plots_metadata_response():
    return {
        "resolutions": ["days", "weeks", "months", "quarters", "years"],
        "backtestStatuses": [
            {
                "training": {"notCompleted": [1, 2, 3, 4, 5, 6, 7]},
                "validation": {"completed": [1], "errored": [2], "notCompleted": [3, 4, 5, 6, 7]},
            },
            {
                "training": {"notCompleted": [1, 2, 3, 4, 5, 6, 7]},
                "validation": {"inProgress": [1, 2, 3, 4, 5, 6, 7]},
            },
        ],
        "holdoutStatuses": {
            "training": {"notSupported": [1, 2, 3, 4, 5, 6, 7]},
            "validation": {"insufficientData": [1, 2, 3, 4, 5, 6, 7]},
        },
        "backtestMetadata": [
            {
                "training": {"startDate": None, "endDate": None},
                "validation": {
                    "startDate": "2014-02-01T00:00:00.000000Z",
                    "endDate": "2014-04-09T00:00:00.000000Z",
                },
            },
            {
                "training": {"startDate": None, "endDate": None},
                "validation": {"startDate": None, "endDate": None},
            },
        ],
        "holdoutMetadata": {
            "training": {"startDate": None, "endDate": None},
            "validation": {"startDate": None, "endDate": None},
        },
    }


@pytest.fixture
def forecast_vs_actual_plot_response():
    return {
        "startDate": "2014-02-01T00:00:00.000000Z",
        "endDate": "2014-04-09T00:00:00.000000Z",
        "resolution": "days",
        "forecastDistances": [1],
        "bins": [
            {
                "actual": 56363,
                "frequency": 10.0,
                "forecasts": [58441],
                "startDate": "2014-02-01T00:00:00.000000Z",
                "endDate": "2014-04-09T00:00:00.000000Z",
                "error": 15024,
                "normalizedError": 0.56,
            },
        ],
        "calendarEvents": [
            {
                "name": "Washington's Birthday",
                "date": "2014-02-17T00:00:00.000000Z",
                "seriesId": None,
            }
        ],
    }


@pytest.fixture
def anomaly_over_time_plots_metadata_response(accuracy_over_time_plots_metadata_response):
    response = accuracy_over_time_plots_metadata_response
    response.pop("forecastDistance")
    return response


@pytest.fixture
def anomaly_over_time_plot_response():
    return {
        "startDate": "2014-02-01T00:00:00.000000Z",
        "endDate": "2014-02-03T00:00:00.000000Z",
        "resolution": "days",
        "bins": [
            {
                "startDate": "2014-02-01T00:00:00.000000Z",
                "frequency": 10.0,
                "predicted": 1.12,
                "endDate": "2014-02-02T00:00:00.000000Z",
            },
            {
                "startDate": "2014-02-02T00:00:00.000000Z",
                "frequency": 0.0,
                "predicted": None,
                "endDate": "2014-02-03T00:00:00.000000Z",
            },
        ],
        "calendarEvents": [
            {
                "date": "2014-02-17T00:00:00.000000Z",
                "seriesId": None,
                "name": "Washington's Birthday",
            }
        ],
    }


@pytest.fixture
def anomaly_over_time_plot_preview_response():
    return {
        "startDate": "1970-01-01T10:30:01.000000Z",
        "endDate": "1970-01-01T11:30:01.000000Z",
        "predictionThreshold": 0.015,
        "bins": [
            {"startDate": "1970-01-01T10:33:01.000000Z", "endDate": "1970-01-01T10:36:01.000000Z"},
            {"startDate": "1970-01-01T10:36:01.000000Z", "endDate": "1970-01-01T10:39:01.000000Z"},
        ],
    }


@responses.activate
def test_get_accuracy_over_time_plots_metadata(
    accuracy_over_time_plots_metadata_response, unittest_endpoint, project_id, model_id
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/accuracyOverTimePlots/metadata/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(accuracy_over_time_plots_metadata_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    metadata = model.get_accuracy_over_time_plots_metadata(forecast_distance=1)
    assert metadata.project_id == project_id
    assert metadata.model_id == model_id
    assert metadata.resolutions == accuracy_over_time_plots_metadata_response["resolutions"]
    assert (
        metadata.forecast_distance == accuracy_over_time_plots_metadata_response["forecastDistance"]
    )
    assert metadata.backtest_statuses[0] == {
        "training": "errored",
        "validation": "insufficientData",
    }
    assert metadata.backtest_statuses[1] == {"training": "notCompleted", "validation": "completed"}
    assert metadata.backtest_metadata[1]["training"] == {"end_date": None, "start_date": None}
    assert metadata.backtest_metadata[1]["validation"] == {
        "end_date": datetime(2014, 2, 1, tzinfo=tzutc()),
        "start_date": datetime(2013, 11, 26, tzinfo=tzutc()),
    }

    assert metadata.holdout_statuses == {"training": "notCompleted", "validation": "completed"}
    assert metadata.holdout_metadata["validation"] == {
        "end_date": datetime(2014, 6, 15, tzinfo=tzutc()),
        "start_date": datetime(2014, 4, 9, tzinfo=tzutc()),
    }

    assert metadata._get_status(backtest=1) == "completed"
    assert metadata._get_status(backtest=1, source="training") == "notCompleted"
    assert metadata._get_status(backtest="holdout", source="training") == "notCompleted"
    assert metadata._get_status(backtest="holdout", source="validation") == "completed"

    assert metadata._get_status(backtest="holdout!") is None
    assert metadata._get_status(backtest=10) is None
    assert metadata._get_status(source="validation!") is None


@responses.activate
def test_get_accuracy_over_time_plot(
    accuracy_over_time_plot_response, unittest_endpoint, project_id, model_id
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/accuracyOverTimePlots/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(accuracy_over_time_plot_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    plot = model.get_accuracy_over_time_plot(
        start_date=datetime(2014, 2, 2), end_date=datetime(2014, 2, 3), max_wait=None
    )
    assert plot.project_id == project_id
    assert plot.model_id == model_id
    assert plot.resolution == accuracy_over_time_plot_response["resolution"]
    assert plot.bins[0] == {
        "actual": 1.11,
        "end_date": datetime(2014, 2, 2, tzinfo=tzutc()),
        "frequency": 10,
        "predicted": 1.12,
        "start_date": datetime(2014, 2, 1, tzinfo=tzutc()),
    }
    assert plot.calendar_events[0] == {
        "date": datetime(2014, 2, 17, tzinfo=tzutc()),
        "name": "Washington's Birthday",
        "series_id": None,
    }
    assert plot.statistics["durbin_watson"] == 0.5


@responses.activate
@pytest.mark.parametrize("plot", ["accuracyOverTime", "forecastVsActual"])
def test_get_datetime_trend_plot_preview(
    datetime_trend_plot_preview_response, unittest_endpoint, project_id, model_id, plot
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/{}Plots/preview/".format(
            unittest_endpoint, project_id, model_id, plot
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(datetime_trend_plot_preview_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    plot = (
        model.get_accuracy_over_time_plot_preview(max_wait=None)
        if "accuracy" in plot
        else model.get_forecast_vs_actual_plot_preview(max_wait=None)
    )
    assert plot.project_id == project_id
    assert plot.model_id == model_id
    assert plot.start_date == datetime(2014, 2, 1, tzinfo=tzutc())
    assert plot.bins[0] == {
        "actual": 100,
        "end_date": datetime(2014, 2, 2, tzinfo=tzutc()),
        "predicted": 101,
        "start_date": datetime(2014, 2, 1, tzinfo=tzutc()),
    }


@responses.activate
def test_get_forecast_vs_actual_plots_metadata(
    forecast_vs_actual_plots_metadata_response, unittest_endpoint, project_id, model_id
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/forecastVsActualPlots/metadata/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(forecast_vs_actual_plots_metadata_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    metadata = model.get_forecast_vs_actual_plots_metadata()
    assert metadata.project_id == project_id
    assert metadata.model_id == model_id
    assert metadata.resolutions == forecast_vs_actual_plots_metadata_response["resolutions"]
    assert metadata.backtest_statuses == [
        {
            "training": {"notCompleted": [1, 2, 3, 4, 5, 6, 7]},
            "validation": {"completed": [1], "notCompleted": [3, 4, 5, 6, 7], "errored": [2]},
        },
        {
            "training": {"notCompleted": [1, 2, 3, 4, 5, 6, 7]},
            "validation": {"inProgress": [1, 2, 3, 4, 5, 6, 7]},
        },
    ]
    assert metadata.backtest_metadata[0]["validation"] == {
        "end_date": datetime(2014, 4, 9, tzinfo=tzutc()),
        "start_date": datetime(2014, 2, 1, tzinfo=tzutc()),
    }
    assert metadata.holdout_statuses == {
        "training": {"notSupported": [1, 2, 3, 4, 5, 6, 7]},
        "validation": {"insufficientData": [1, 2, 3, 4, 5, 6, 7]},
    }

    assert metadata._get_status(backtest=1) == {"inProgress": [1, 2, 3, 4, 5, 6, 7]}
    assert metadata._get_status(backtest=1, source="training") == {
        "notCompleted": [1, 2, 3, 4, 5, 6, 7]
    }
    assert metadata._get_status(backtest="holdout", source="training") == {
        "notSupported": [1, 2, 3, 4, 5, 6, 7]
    }
    assert metadata._get_status(backtest="holdout", source="validation") == {
        "insufficientData": [1, 2, 3, 4, 5, 6, 7]
    }


@responses.activate
def test_get_forecast_vs_actual_plot(
    forecast_vs_actual_plot_response, unittest_endpoint, project_id, model_id
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/forecastVsActualPlots/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(forecast_vs_actual_plot_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    plot = model.get_forecast_vs_actual_plot(
        start_date=datetime(2014, 2, 2), end_date=datetime(2014, 2, 3), max_wait=None
    )
    assert plot.project_id == project_id
    assert plot.model_id == model_id
    assert plot.resolution == forecast_vs_actual_plot_response["resolution"]
    assert plot.forecast_distances == forecast_vs_actual_plot_response["forecastDistances"]
    assert plot.bins[0] == {
        "actual": 56363,
        "frequency": 10.0,
        "forecasts": [58441],
        "start_date": datetime(2014, 2, 1, tzinfo=tzutc()),
        "end_date": datetime(2014, 4, 9, tzinfo=tzutc()),
        "error": 15024,
        "normalized_error": 0.56,
    }
    assert plot.calendar_events[0] == {
        "date": datetime(2014, 2, 17, tzinfo=tzutc()),
        "name": "Washington's Birthday",
        "series_id": None,
    }


@responses.activate
def test_get_anomaly_over_time_plots_metadata(
    anomaly_over_time_plots_metadata_response, unittest_endpoint, project_id, model_id
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/anomalyOverTimePlots/metadata/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(anomaly_over_time_plots_metadata_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    metadata = model.get_anomaly_over_time_plots_metadata()
    assert metadata.project_id == project_id
    assert metadata.model_id == model_id
    assert metadata.resolutions == anomaly_over_time_plots_metadata_response["resolutions"]
    assert metadata.backtest_statuses[0] == {
        "training": "errored",
        "validation": "insufficientData",
    }
    assert metadata.backtest_statuses[1] == {"training": "notCompleted", "validation": "completed"}
    assert metadata.backtest_metadata[1]["training"] == {"end_date": None, "start_date": None}
    assert metadata.backtest_metadata[1]["validation"] == {
        "end_date": datetime(2014, 2, 1, tzinfo=tzutc()),
        "start_date": datetime(2013, 11, 26, tzinfo=tzutc()),
    }


@responses.activate
def test_get_anomaly_over_time_plot(
    anomaly_over_time_plot_response, unittest_endpoint, project_id, model_id
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/anomalyOverTimePlots/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(anomaly_over_time_plot_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    plot = model.get_anomaly_over_time_plot(
        start_date=datetime(2014, 2, 2), end_date=datetime(2014, 2, 3), max_wait=None
    )
    assert plot.project_id == project_id
    assert plot.model_id == model_id
    assert plot.resolution == anomaly_over_time_plot_response["resolution"]
    assert plot.bins[0] == {
        "end_date": datetime(2014, 2, 2, tzinfo=tzutc()),
        "frequency": 10,
        "predicted": 1.12,
        "start_date": datetime(2014, 2, 1, tzinfo=tzutc()),
    }
    assert plot.calendar_events[0] == {
        "date": datetime(2014, 2, 17, tzinfo=tzutc()),
        "name": "Washington's Birthday",
        "series_id": None,
    }


@responses.activate
def test_get_anomaly_over_time_plot_preview(
    anomaly_over_time_plot_preview_response, unittest_endpoint, project_id, model_id
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/anomalyOverTimePlots/preview/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(anomaly_over_time_plot_preview_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    plot = model.get_anomaly_over_time_plot_preview(max_wait=None)

    assert plot.prediction_threshold == 0.015
    assert plot.project_id == project_id
    assert plot.model_id == model_id
    assert plot.start_date == datetime(1970, 1, 1, 10, 30, 1, tzinfo=tzutc())
    assert plot.bins[0] == {
        "end_date": datetime(1970, 1, 1, 10, 36, 1, tzinfo=tzutc()),
        "start_date": datetime(1970, 1, 1, 10, 33, 1, tzinfo=tzutc()),
    }


def _add_compute_route_to_responses(unittest_endpoint, project_id, model_id, responses):
    compute_url = "{}/projects/{}/datetimeModels/{}/datetimeTrendPlots/".format(
        unittest_endpoint, project_id, model_id
    )
    job_url = "{}/projects/{}/jobs/1/".format(unittest_endpoint, project_id)

    responses.add(responses.POST, compute_url, status=202, headers={"Location": job_url})
    responses.add(
        responses.GET,
        job_url,
        status=303,
        content_type="application/json",
        body=json.dumps(
            {
                "status": "COMPLETED",
                "model_id": model_id,
                "is_blocked": False,
                "url": job_url,
                "job_type": "datetime_trend_plots",
                "project_id": project_id,
                "id": 1,
            }
        ),
        headers={"Location": "{}result/".format(job_url)},
    )


@responses.activate
def test_accuracy_over_time_plots_auto_compute_before_retrieval(
    unittest_endpoint, project_id, model_id, accuracy_over_time_plots_metadata_response,
):
    def reset_responses():
        responses.reset()
        _add_compute_route_to_responses(unittest_endpoint, project_id, model_id, responses)
        responses.add(
            responses.GET,
            "{}/projects/{}/datetimeModels/{}/accuracyOverTimePlots/metadata/".format(
                unittest_endpoint, project_id, model_id
            ),
            status=200,
            content_type="application/json",
            body=json.dumps(accuracy_over_time_plots_metadata_response),
        )

    model = DatetimeModel(id=model_id, project_id=project_id)

    # Backtest 1, Training, FD 10: Status is not completed. Run compute job
    reset_responses()
    model._compute_accuracy_over_time_plot_if_not_computed(
        backtest=1, source="training", forecast_distance=10, max_wait=100
    )
    assert request_body_to_json(responses.calls[1].request) == {
        "backtest": 1,
        "source": "training",
        "forecastDistanceStart": 10,
        "forecastDistanceEnd": 10,
    }

    # Backtest 1, Training, FD 10: Status is completed. Do not run compute job
    reset_responses()
    model._compute_accuracy_over_time_plot_if_not_computed(
        backtest=0, source="training", forecast_distance=10, max_wait=100
    )
    assert len(responses.calls) == 1 and responses.calls[0].request.method == responses.GET


@responses.activate
def test_forecast_vs_actual_plots_auto_compute_before_retrieval(
    unittest_endpoint, project_id, model_id, forecast_vs_actual_plots_metadata_response,
):
    def reset_responses():
        responses.reset()
        _add_compute_route_to_responses(unittest_endpoint, project_id, model_id, responses)
        responses.add(
            responses.GET,
            "{}/projects/{}/datetimeModels/{}/forecastVsActualPlots/metadata/".format(
                unittest_endpoint, project_id, model_id
            ),
            status=200,
            content_type="application/json",
            body=json.dumps(forecast_vs_actual_plots_metadata_response),
        )

    model = DatetimeModel(id=model_id, project_id=project_id)

    # Backtest 0, Validation: FD 2 is errored, but FD 3 and FD 4 are not completed. Run compute job
    reset_responses()
    model._compute_forecast_vs_actual_plot_if_not_computed(
        backtest=0,
        source="validation",
        forecast_distance_start=2,
        forecast_distance_end=4,
        max_wait=100,
    )
    assert request_body_to_json(responses.calls[1].request) == {
        "backtest": 0,
        "source": "validation",
        "forecastDistanceStart": 2,
        "forecastDistanceEnd": 4,
    }

    # Backtest 0, Validation: FD 1 is completed, FD 2 is errored. Do not run compute job
    reset_responses()
    model._compute_forecast_vs_actual_plot_if_not_computed(
        backtest=0,
        source="validation",
        forecast_distance_start=1,
        forecast_distance_end=2,
        max_wait=100,
    )
    assert len(responses.calls) == 1 and responses.calls[0].request.method == responses.GET

    # Backtest 1, Validation: All FDs are in progress. Do not run compute job
    reset_responses()
    model._compute_forecast_vs_actual_plot_if_not_computed(
        backtest=1,
        source="validation",
        forecast_distance_start=None,
        forecast_distance_end=None,
        max_wait=100,
    )
    assert len(responses.calls) == 1 and responses.calls[0].request.method == responses.GET

    # Backtest 0, Validation: FD 1 is completed, FD 2 is errored, but FD 3-4 are not completed.
    # Run compute job
    reset_responses()
    model._compute_forecast_vs_actual_plot_if_not_computed(
        backtest=0,
        source="validation",
        forecast_distance_start=None,
        forecast_distance_end=4,
        max_wait=100,
    )
    assert request_body_to_json(responses.calls[1].request) == {
        "backtest": 0,
        "source": "validation",
        "forecastDistanceEnd": 4,
    }

    # Backtest 0, Validation: FD 3-7 are not completed. Run compute job
    reset_responses()
    model._compute_forecast_vs_actual_plot_if_not_computed(
        backtest=0,
        source="validation",
        forecast_distance_start=3,
        forecast_distance_end=None,
        max_wait=100,
    )
    assert request_body_to_json(responses.calls[1].request) == {
        "backtest": 0,
        "source": "validation",
        "forecastDistanceStart": 3,
    }

    # Backtest 0, Validation: FD 1 is completed, FD 2 is errored, FD 3-7 are not completed.
    # Run compute job
    reset_responses()
    model._compute_forecast_vs_actual_plot_if_not_computed(
        backtest=0,
        source="validation",
        forecast_distance_start=None,
        forecast_distance_end=None,
        max_wait=100,
    )

    assert request_body_to_json(responses.calls[1].request) == {
        "backtest": 0,
        "source": "validation",
    }


@responses.activate
def test_anomaly_over_time_plots_auto_compute_before_retrieval(
    unittest_endpoint, project_id, model_id, anomaly_over_time_plots_metadata_response,
):
    def reset_responses():
        responses.reset()
        _add_compute_route_to_responses(unittest_endpoint, project_id, model_id, responses)
        responses.add(
            responses.GET,
            "{}/projects/{}/datetimeModels/{}/anomalyOverTimePlots/metadata/".format(
                unittest_endpoint, project_id, model_id
            ),
            status=200,
            content_type="application/json",
            body=json.dumps(anomaly_over_time_plots_metadata_response),
        )

    model = DatetimeModel(id=model_id, project_id=project_id)

    # Backtest 1, Training: Status is not completed. Run compute job
    reset_responses()
    model._compute_anomaly_over_time_plot_if_not_computed(
        backtest=1, source="training", max_wait=100
    )
    assert request_body_to_json(responses.calls[1].request) == {
        "backtest": 1,
        "source": "training",
    }

    # Backtest 1, Training: Status is completed. Do not run compute job
    reset_responses()
    model._compute_anomaly_over_time_plot_if_not_computed(
        backtest=0, source="training", max_wait=100
    )
    assert len(responses.calls) == 1 and responses.calls[0].request.method == responses.GET



import pytz


class DeploymentQueryBuilderMixin(object):
    @staticmethod
    def _build_query_params(start_time=None, end_time=None, **kwargs):
        def timezone_aware(dt):
            return dt.replace(tzinfo=pytz.utc) if not dt.tzinfo else dt

        if start_time:
            kwargs["start"] = timezone_aware(start_time).isoformat()
        if end_time:
            kwargs["end"] = timezone_aware(end_time).isoformat()
        kwargs = {key: value for key, value in kwargs.items() if value is not None}
        return kwargs



import json

import pytest
import responses
import trafaret

from datarobot import errors
from datarobot.models.training_predictions import RowsIterator


@responses.activate
def test_rows_iterator__empty_response__ok(client):
    body = {"data": [], "next": None}
    responses.add(
        responses.GET,
        "https://host_name.com/foobar",
        body=json.dumps(body),
        content_type="application/json",
    )

    it = RowsIterator(client, "foobar")

    assert list(it) == []


@responses.activate
def test_rows_iterator__two_items__ok(client):
    url = "https://host_name.com/foobar"
    body = {"data": [{"foo": "bar"}], "next": "{}?offset=1&limit=1".format(url)}
    responses.add(
        responses.GET, url, body=json.dumps(body), content_type="application/json",
    )

    it = RowsIterator(client, "foobar", limit=1)

    assert next(it, None) == {"foo": "bar"}
    assert len(responses.calls) == 1

    responses.reset()
    body2 = {"data": [{"bar": "foo"}], "next": "{}?offset=2&limit=1".format(url)}
    responses.add(
        responses.GET, url, body=json.dumps(body2), content_type="application/json",
    )

    assert next(it, None) == {"bar": "foo"}
    assert len(responses.calls) == 1

    responses.reset()
    body3 = {"data": [], "next": "{}?offset=3&limit=1".format(url)}
    responses.add(
        responses.GET, url, body=json.dumps(body3), content_type="application/json",
    )

    assert next(it, None) is None
    assert len(responses.calls) == 1


@responses.activate
def test_rows_iterator__http_504__raise_error(client):
    url = "https://host_name.com/foobar"
    body = {"message": "test-error"}
    responses.add(
        responses.GET, url, body=json.dumps(body), content_type="application/json", status=504,
    )

    with pytest.raises(errors.ServerError) as exc:
        it = RowsIterator(client, "foobar")
        next(it, None)

    assert exc.value.status_code == 504


@responses.activate
def test_rows_iterator__no_data_field__raise_error(client):
    url = "https://host_name.com/foobar"
    body = {"message": "test-error", "next": "test"}
    responses.add(
        responses.GET, url, body=json.dumps(body), content_type="application/json",
    )

    with pytest.raises(Exception) as exc:
        it = RowsIterator(client, "foobar")
        next(it, None)

    assert isinstance(exc.value, trafaret.DataError)



#! /usr/bin/env python3
"""Display a single image in a project.

The image id must be known before using this code. That may be obtained
through getting sample iamges, activation maps, or embeddings. See
examples in the same directory as this file to see how that is done.
"""
import io

import PIL.Image

from datarobot.models import Project
from datarobot.models.visualai import Image

image_id = "5e7e562528513130ab237875"
project = Project.list(search_params={"project_name": "dataset_2k.zip"})[0]
image = Image.get(project.id, image_id)
print(image)
with io.BytesIO(image.image_bytes) as bio, PIL.Image.open(bio) as img:
    img.show()



import json

import mock
import pytest
import responses

from datarobot import Blueprint, errors, Job, ModelJob, Project
from datarobot.models.modeljob import wait_for_async_model_creation
from tests.utils import assert_raised_regex, SDKTestcase


class TestWaitForAsyncModelCreation(SDKTestcase):
    mock_model_job_data = {
        "status": "inprogress",
        "processes": [
            "One-Hot Encoding",
            "Bernoulli Naive Bayes classifier (scikit-learn)",
            "Missing Values Imputed",
            "Gaussian Naive Bayes classifier (scikit-learn)",
            "Naive Bayes combiner classifier",
            "Calibrate predictions",
        ],
        "projectId": "p-id",
        "samplePct": 28.349,
        "modelType": "Naive Bayes combiner classifier",
        "featurelistId": "56d8620bccf94e26f37af0a3",
        "modelCategory": "model",
        "blueprintId": "2a1b9ae97fe61880332e196c770c1f9f",
        "isBlocked": False,
        "id": "1",
    }

    def setUp(self):
        super(TestWaitForAsyncModelCreation, self).setUp()
        self.pid = "p-id"
        self.model_job_id = "5"
        self.get_job_url = "https://host_name.com/projects/{}/modelJobs/{}/".format(
            self.pid, self.model_job_id,
        )

    @responses.activate
    def test_timeout(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(self.mock_model_job_data),
            content_type="application/json",
        )
        with pytest.raises(errors.AsyncTimeoutError) as exc_info:
            wait_for_async_model_creation(self.pid, self.model_job_id, max_wait=1)
        assert_raised_regex(exc_info, "Model creation timed out in")

    @responses.activate
    @mock.patch("datarobot.models.modeljob.ModelJob.get_model")
    def test_success(self, get_model):
        responses.add(
            responses.GET, self.get_job_url, status=303, body="", content_type="application/json"
        )
        wait_for_async_model_creation(self.pid, self.model_job_id)
        self.assertEqual(get_model.call_count, 1)

    @responses.activate
    @mock.patch("datarobot.models.modeljob.ModelJob.get_model")
    def test_error(self, get_model):
        data = dict(self.mock_model_job_data, status="error")
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(data),
            content_type="application/json",
        )

        with pytest.raises(errors.AsyncModelCreationError) as exc_info:
            wait_for_async_model_creation(self.pid, self.model_job_id)
        assert_raised_regex(exc_info, "Model creation unsuccessful")


class TestJobModel(SDKTestcase):
    fully_trained_model = """
            {
        "featurelistId": "556cdfbd100d2b10048c7941",
        "processes": ["One", "Two", "Three"],
        "featurelistName": "Informative Features",
        "projectId": "p-id",
        "samplePct": 64,
        "modelType": "AVG Blender",
        "metrics": {
            "AUC": {
                "holdout": 0.76603,
                "validation": 0.64141,
                "crossValidation": 0.7625240000000001
            },
            "Rate@Top5%": {
                "holdout": 1,
                "validation": 0.5,
                "crossValidation": 0.9
            },
            "Rate@TopTenth%": {
                "holdout": 1,
                "validation": 1,
                "crossValidation": 1
            },
            "RMSE": {
                "holdout": 0.42054,
                "validation": 0.44396,
                "crossValidation": 0.40162000000000003
            },
            "LogLoss": {
                "holdout": 0.53707,
                "validation": 0.58051,
                "crossValidation": 0.5054160000000001
            },
            "FVE Binomial": {
                "holdout": 0.17154,
                "validation": 0.03641,
                "crossValidation": 0.17637399999999998
            },
            "Gini Norm": {
                "holdout": 0.53206,
                "validation": 0.28282,
                "crossValidation": 0.525048
            },
            "Rate@Top10%": {
                "holdout": 1,
                "validation": 0.25,
                "crossValidation": 0.7
            }
        },
        "blueprintId": "a4fd9d17a8ca62ee00590dd704dae6a8",
        "id": "5223deadbeefdeadbeef1234"
    }
    """

    final_model_location = (
        "https://host_name.com/projects/55666eb9100d2b109b59e267/"
        "models/5223deadbeefdeadbeef1234/"
    )

    mock_generic_job_data_inprogress = {
        "status": "inprogress",
        "url": "https://host_name.com/projects/p-id/modelJobs/1/",
        "id": "1",
        "jobType": "model",
        "isBlocked": False,
        "projectId": "p-id",
    }

    mock_generic_job_data_completed = dict(mock_generic_job_data_inprogress, status="COMPLETED")

    mock_model_job_data = {
        "status": "inprogress",
        "processes": [
            "One-Hot Encoding",
            "Bernoulli Naive Bayes classifier (scikit-learn)",
            "Missing Values Imputed",
            "Gaussian Naive Bayes classifier (scikit-learn)",
            "Naive Bayes combiner classifier",
            "Calibrate predictions",
        ],
        "projectId": "p-id",
        "samplePct": 28.349,
        "modelType": "Naive Bayes combiner classifier",
        "featurelistId": "56d8620bccf94e26f37af0a3",
        "modelCategory": "model",
        "blueprintId": "2a1b9ae97fe61880332e196c770c1f9f",
        "isBlocked": False,
        "id": "1",
    }

    mock_predict_job_data = {
        "status": "error",
        "url": "https://host_name.com/projects/p-id/predictJobs/64/",
        "id": "64",
        "jobType": "predict",
        "isBlocked": False,
        "projectId": "p-id",
    }

    def test_instantiate_job(self):
        data = {
            "status": "queue",
            "processes": ["One-Hot Encoding", "Missing Values Imputed", "RuleFit Classifier"],
            "projectId": "556902e8100d2b3728d47551",
            "samplePct": 64,
            "modelType": "RuleFit Classifier",
            "featurelistId": "556902eb100d2b37d1130771",
            "blueprintId": "a8959bc1d46f07fb3dc14db7c1e3fc99",
            "isBlocked": False,
            "id": 11,
            "modelId": "556902ef100d2b37da13077d",
        }
        job = ModelJob(data)
        self.assertEqual(job.status, "queue")
        self.assertEqual(job.job_type, "model")
        self.assertEqual(job.processes, data["processes"])
        self.assertEqual(job.sample_pct, 64)
        self.assertEqual(job.model_type, "RuleFit Classifier")
        assert job.featurelist_id == data["featurelistId"]
        self.assertIsInstance(job.project, Project)
        self.assertEqual(job.project.id, "556902e8100d2b3728d47551")
        self.assertIsInstance(job.blueprint, Blueprint)
        self.assertEqual(job.blueprint.id, "a8959bc1d46f07fb3dc14db7c1e3fc99")

        repr(job)

        with self.assertRaises(ValueError):
            ModelJob("qid")

    def test_future_proof(self):
        ModelJob(dict(self.mock_model_job_data, future="new"))

    @responses.activate
    def test_from_job(self):
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/modelJobs/1/",
            status=200,
            body=json.dumps(self.mock_model_job_data),
            content_type="application/json",
        )
        generic_job = Job(self.mock_generic_job_data_inprogress)
        model_job = ModelJob.from_job(generic_job)

        self.assertEqual(model_job.status, "inprogress")
        self.assertEqual(model_job.job_type, "model")
        self.assertEqual(model_job.project, Project("p-id"))
        self.assertEqual(model_job.processes, self.mock_model_job_data["processes"])

    def test_from_job_wrong_type(self):
        generic_job = Job(self.mock_predict_job_data)
        with self.assertRaises(ValueError):
            ModelJob.from_job(generic_job)

    @responses.activate
    def test_cancel(self):
        data = {
            "status": "inprogress",
            "samplepct": 64.0,
            "processes": [
                "One-Hot Encoding",
                "Missing Values Imputed",
                "Standardize",
                "Linear Regression",
            ],
            "modelType": "Linear Regression",
            "featurelistId": "55666d05100d2b01a1104dae",
            "blueprintId": "3bb4665320be633b30a9601b3e73284d",
            "projectId": "55666eb9100d2b109b59e267",
            "isBlocked": False,
            "id": 5,
            "modelId": "55666d0d100d2b01b1104db4",
        }
        job = ModelJob(data)
        responses.add(
            responses.DELETE,
            "https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/",
            status=204,
            body="",
            content_type="application/json",
        )
        job.cancel()
        self.assertEqual(responses.calls[0].request.method, "DELETE")

    @responses.activate
    def test_get_success(self):
        data = {
            "status": "inprogress",
            "samplepct": 64.0,
            "processes": [
                "One-Hot Encoding",
                "Missing Values Imputed",
                "Standardize",
                "Linear Regression",
            ],
            "modelType": "Linear Regression",
            "featurelistId": "55666d05100d2b01a1104dae",
            "blueprintId": "3bb4665320be633b30a9601b3e73284d",
            "projectId": "55666eb9100d2b109b59e267",
            "isBlocked": False,
            "id": 5,
            "modelId": "55666d0d100d2b01b1104db4",
        }
        responses.add(
            responses.GET,
            "https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/",
            status=200,
            body=json.dumps(data),
            content_type="application/json",
        )
        job = ModelJob.get("55666eb9100d2b109b59e267", 5)
        self.assertEqual(responses.calls[0].request.method, "GET")
        self.assertEqual(job.status, "inprogress")

    @responses.activate
    def test_get_redirect(self):
        responses.add(
            responses.GET,
            "https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/",
            content_type="application/json",
            status=303,
            body="",
            adding_headers={"Location": "http://pam/api/v2/projects/p-id/models/id/"},
        )
        with self.assertRaises(errors.PendingJobFinished):
            ModelJob.get("55666eb9100d2b109b59e267", 5)

    @responses.activate
    def test_get_exceptional(self):
        responses.add(
            responses.GET,
            "https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/",
            status=404,
            body="",
        )
        with self.assertRaises(errors.AppPlatformError):
            ModelJob.get("55666eb9100d2b109b59e267", 5)

    @responses.activate
    def test_get_unexpected_status_code(self):
        responses.add(
            responses.GET,
            "https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/",
            status=202,
            body="",
        )
        with pytest.raises(errors.AsyncFailureError) as exc_info:
            ModelJob.get("55666eb9100d2b109b59e267", 5)
        assert_raised_regex(exc_info, "Server unexpectedly returned status code")

    @responses.activate
    def test_get_model_success(self):
        responses.add(
            responses.GET,
            "https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/",
            status=303,
            body="",
            adding_headers={"Location": self.final_model_location},
        )
        responses.add(
            responses.GET, self.final_model_location, status=200, body=self.fully_trained_model
        )
        ModelJob.get_model("55666eb9100d2b109b59e267", 5)

    @responses.activate
    def test_get_model_from_job_success(self):
        generic_job = Job(self.mock_generic_job_data_inprogress)
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/jobs/1/",
            status=303,
            body=json.dumps(self.mock_generic_job_data_completed),
            adding_headers={"Location": self.final_model_location},
        )
        responses.add(
            responses.GET, self.final_model_location, status=200, body=self.fully_trained_model
        )
        generic_job.get_result()

    @responses.activate
    def test_get_model_from_job_not_finished(self):
        generic_job = Job(self.mock_generic_job_data_inprogress)
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/jobs/1/",
            status=200,
            body=json.dumps(self.mock_generic_job_data_inprogress),
            adding_headers={"Location": self.final_model_location},
        )
        with self.assertRaises(errors.JobNotFinished):
            generic_job.get_result()

    @responses.activate
    def test_get_model_not_finished(self):
        model_job_data = {
            "status": "queue",
            "processes": ["One-Hot Encoding", "Missing Values Imputed", "RuleFit Classifier"],
            "projectId": "556902e8100d2b3728d47551",
            "samplePct": 64,
            "modelType": "RuleFit Classifier",
            "featurelistId": "556902eb100d2b37d1130771",
            "blueprintId": "a8959bc1d46f07fb3dc14db7c1e3fc99",
            "isBlocked": False,
            "id": 5,
            "modelId": "556902ef100d2b37da13077d",
        }

        responses.add(
            responses.GET,
            "https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/",
            status=200,
            body=json.dumps(model_job_data),
        )
        with self.assertRaises(errors.JobNotFinished):
            ModelJob.get_model("55666eb9100d2b109b59e267", 5)

    @responses.activate
    def test_get_model_exceptional(self):
        responses.add(
            responses.GET,
            "https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/",
            status=404,
            body="",
        )
        with self.assertRaises(errors.AppPlatformError):
            ModelJob.get_model("55666eb9100d2b109b59e267", 5)

    @responses.activate
    def test_get_model_unexpected_status_code(self):
        final_model_location = (
            "https://host_name.com/projects/55666eb9100d2b109b59e267/"
            "models/5223deadbeefdeadbeef1234/"
        )

        responses.add(
            responses.GET,
            "https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/",
            status=202,
            body="",
            adding_headers={"Location": final_model_location},
        )

        with pytest.raises(errors.AsyncFailureError) as exc_info:
            ModelJob.get_model("55666eb9100d2b109b59e267", 5)
        assert_raised_regex(exc_info, "Server unexpectedly returned status code")



from datarobot.models.api_object import APIObject
from datarobot.utils import deprecation_warning, get_id_from_response


class ComplianceDocumentation(APIObject):
    """
    A :ref:`compliance documentation <compliance_documentation_overview>` object.

    .. versionadded:: v2.14

    Attributes
    ----------
    project_id : str
        the id of the project
    model_id : str
        the id of the model
    template_id : str or None
        optional id of the template for the generated doc. See documentation for
        :py:class:`ComplianceDocTemplate
        <datarobot.models.compliance_doc_template.ComplianceDocTemplate>` for more info.

    Examples
    --------
    .. code-block:: python

        doc = ComplianceDocumentation('project-id', 'model-id')
        job = doc.generate()
        job.wait_for_completion()
        doc.download('example.docx')

    """

    def __init__(self, project_id, model_id, template_id=None):
        self.project_id = project_id
        self.model_id = model_id
        self.template_id = template_id

        deprecation_warning(
            "ComplianceDocumentation",
            deprecated_since_version="v2.24",
            will_remove_version="v2.27",
            message="Use AutomatedDocument instead.",
        )

    def generate(self):

        """
        Start a job generating model compliance documentation.

        Returns
        -------
        Job
            an instance of an async job
        """
        from . import Job

        url = "projects/{}/models/{}/complianceDocs/".format(self.project_id, self.model_id)
        payload = {"template_id": self.template_id} if self.template_id else {}
        response = self._client.post(url, data=payload)
        job_id = get_id_from_response(response)

        return Job.get(self.project_id, job_id)

    def download(self, filepath):
        """ Download the generated compliance documentation file and save it
        to the specified path. The generated file has a DOCX format.

        Parameters
        ----------
        filepath : str
            A file path, e.g. "/path/to/save/compliance_documentation.docx"
        """
        url = "projects/{}/models/{}/complianceDocs/".format(self.project_id, self.model_id)
        response = self._client.get(url, stream=True)
        with open(filepath, "wb") as f:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                f.write(chunk)



import dateutil
import trafaret as t

from datarobot.models.api_object import APIObject

from ..enums import DATA_DRIFT_METRIC
from ..helpers.deployment_monitoring import DeploymentQueryBuilderMixin
from ..utils import encode_utf8_if_py2, from_api


class TargetDrift(APIObject, DeploymentQueryBuilderMixin):
    """Deployment target drift information.

    Attributes
    ----------
    model_id : str
        the model used to retrieve target drift metric
    period : dict
        the time period used to retrieve target drift metric
    metric : str
        the data drift metric
    target_name : str
        name of the target
    drift_score : float
        target drift score
    sample_size : int
        count of data points for comparison
    baseline_sample_size : int
        count of data points for baseline
    """

    _path = "deployments/{}/targetDrift/"
    _period = t.Dict(
        {
            t.Key("start"): t.String >> (lambda s: dateutil.parser.parse(s)),
            t.Key("end"): t.String >> (lambda s: dateutil.parser.parse(s)),
        }
    )
    _converter = t.Dict(
        {
            t.Key("period"): _period,
            t.Key("metric"): t.Enum(*DATA_DRIFT_METRIC.ALL) | t.Null,
            t.Key("model_id"): t.String() | t.Null,
            t.Key("target_name"): t.String() | t.Null,
            t.Key("drift_score"): t.Float() | t.Null,
            t.Key("sample_size"): t.Int() | t.Null,
            t.Key("baseline_sample_size"): t.Int() | t.Null,
        }
    ).allow_extra("*")

    def __init__(
        self,
        period=None,
        metric=None,
        model_id=None,
        target_name=None,
        drift_score=None,
        sample_size=None,
        baseline_sample_size=None,
    ):
        self.period = period or {}
        self.metric = metric
        self.model_id = model_id
        self.target_name = target_name
        self.drift_score = drift_score
        self.sample_size = sample_size
        self.baseline_sample_size = baseline_sample_size

    def __repr__(self):
        return encode_utf8_if_py2(
            u"{}({} | {} | {} - {})".format(
                self.__class__.__name__,
                self.model_id,
                self.target_name,
                self.period.get("start"),
                self.period.get("end"),
            )
        )

    @classmethod
    def get(cls, deployment_id, model_id=None, start_time=None, end_time=None, metric=None):
        """Retrieve target drift information over a certain time period.

        .. versionadded:: v2.21

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        model_id : str
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        metric : str
            (New in version v2.22) metric used to calculate the drift score

        Returns
        -------
        target_drift : TargetDrift
            the queried target drift information

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment, TargetDrift
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            target_drift = TargetDrift.get(deployment.id)
            target_drift.period['end']
            >>>'2019-08-01 00:00:00+00:00'
            target_drift.drift_score
            >>>0.03423
            accuracy.target_name
            >>>'readmitted'
        """

        path = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time, end_time=end_time, model_id=model_id
        )
        if metric:
            params["metric"] = metric
        data = cls._client.get(path, params=params).json()
        data = from_api(data, keep_null_keys=True)
        return cls.from_data(data)


class FeatureDrift(APIObject, DeploymentQueryBuilderMixin):
    """Deployment feature drift information.

    Attributes
    ----------
    model_id : str
        the model used to retrieve feature drift metric
    period : dict
        the time period used to retrieve feature drift metric
    metric : str
        the data drift metric
    name : str
        name of the feature
    drift_score : float
        feature drift score
    sample_size : int
        count of data points for comparison
    baseline_sample_size : int
        count of data points for baseline
    """

    _path = "deployments/{}/featureDrift/"
    _period = t.Dict(
        {
            t.Key("start"): t.String >> (lambda s: dateutil.parser.parse(s)),
            t.Key("end"): t.String >> (lambda s: dateutil.parser.parse(s)),
        }
    )
    _converter = t.Dict(
        {
            t.Key("period"): _period,
            t.Key("metric"): t.Enum(*DATA_DRIFT_METRIC.ALL) | t.Null,
            t.Key("model_id"): t.String() | t.Null,
            t.Key("name"): t.String() | t.Null,
            t.Key("drift_score"): t.Float() | t.Null,
            t.Key("feature_impact"): t.Float() | t.Null,
            t.Key("sample_size"): t.Int() | t.Null,
            t.Key("baseline_sample_size"): t.Int() | t.Null,
        }
    ).allow_extra("*")

    def __init__(
        self,
        period=None,
        metric=None,
        model_id=None,
        name=None,
        drift_score=None,
        feature_impact=None,
        sample_size=None,
        baseline_sample_size=None,
    ):
        self.period = period or {}
        self.metric = metric
        self.model_id = model_id
        self.name = name
        self.drift_score = drift_score
        self.feature_impact = feature_impact
        self.sample_size = sample_size
        self.baseline_sample_size = baseline_sample_size

    def __repr__(self):
        return encode_utf8_if_py2(
            u"{}({} | {} | {} - {})".format(
                self.__class__.__name__,
                self.model_id,
                self.name,
                self.period.get("start"),
                self.period.get("end"),
            )
        )

    @classmethod
    def list(cls, deployment_id, model_id=None, start_time=None, end_time=None, metric=None):
        """Retrieve drift information for deployment's features over a certain time period.

        .. versionadded:: v2.21

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        model_id : str
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        metric : str
            (New in version v2.22) metric used to calculate the drift score

        Returns
        -------
        feature_drift_data : [FeatureDrift]
            the queried feature drift information

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment, TargetDrift
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            feature_drift = FeatureDrift.list(deployment.id)[0]
            feature_drift.period
            >>>'2019-08-01 00:00:00+00:00'
            feature_drift.drift_score
            >>>0.252
            feature_drift.name
            >>>'age'
        """

        url = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time, end_time=end_time, model_id=model_id
        )
        if metric:
            params["metric"] = metric
        response_json = cls._client.get(url, params=params).json()
        response_json = from_api(response_json, keep_null_keys=True)

        period = response_json.get("period", {})
        metric = response_json.get("metric")
        model_id = response_json.get("model_id")

        def _from_data_item(item):
            item["period"] = period
            item["metric"] = metric
            item["model_id"] = model_id
            return cls.from_data(item)

        data = []
        for item in response_json["data"]:
            data.append(_from_data_item(item))
        while response_json["next"] is not None:
            response_json = cls._client.get(response_json["next"]).json()
            response_json = from_api(response_json, keep_null_keys=True)
            for item in response_json["data"]:
                data.append(_from_data_item(item))

        return data



import os
import tempfile
import unittest

import mock
import pytest
import responses
import six

import datarobot as dr
from datarobot.client import Client, get_client, set_client
from datarobot.enums import DEFAULT_TIMEOUT
from datarobot.errors import AppPlatformError, ClientError
from datarobot.rest import RESTClientObject, Retry
from tests.utils import SDKTestcase

if six.PY3:
    from unittest.mock import patch
else:
    from mock import patch


@pytest.mark.no_client
class TestClient(object):
    @pytest.mark.usefixtures("no_client_version_check")
    def test_instantiation(self):
        """
        Basic client installation.
        """
        client = Client(token="t-token", endpoint="https://host_name.com")
        assert get_client() is client

    @pytest.mark.usefixtures("no_client_version_check")
    @mock.patch("datarobot.client._file_exists", return_value=False)
    def test_instantiation_without_env(self, _):
        """
        Basic client installation by get_client without configuration.
        """
        with mock.patch("os.environ", {"DATAROBOT_ENDPOINT": "https://host_name.com"}):
            with pytest.raises(ValueError):
                get_client()

    @pytest.mark.usefixtures("no_client_version_check")
    @mock.patch("datarobot.client._file_exists")
    def test_no_endpoint_fails(self, mock_file_exists):
        mock_file_exists.return_value = False
        with mock.patch("os.environ", {}):
            with pytest.raises(ValueError):
                Client(token="NOTATOKEN")

    @pytest.mark.usefixtures("no_client_version_check")
    def test_token_and_endpoint_is_okay(self, no_client_version_check):
        Client(token="token", endpoint="https://need_an_endpoint.com")
        assert no_client_version_check.called

    def test_bad_endpoint_fails(self):
        """
        Client with bad endpoint
        """
        with mock.patch("datarobot.rest.RESTClientObject.get") as mock_version_check:
            mock_version_check.side_effect = ClientError("dummy msg", 12345)
            with pytest.raises(UserWarning, match="Server did not reply"):
                Client(token="t-token", endpoint="https://host_name.com")

    @pytest.mark.usefixtures("no_client_version_check")
    def test_re_instantiation(self):
        """
        Client re installation.
        """
        previous = Client(token="t-****", endpoint="https://end.com")
        old_client = set_client(
            RESTClientObject(auth=("u-**********", "p-******"), endpoint="https://host_name.com")
        )
        assert previous is old_client

    @pytest.mark.usefixtures("no_client_version_check")
    @responses.activate
    def test_recognizing_domain_on_instance(self):
        raw = """{"api_token": "some_token"}"""
        responses.add(
            responses.POST,
            "https://host_name.com/api/v2/api_token/",
            body=raw,
            status=201,
            content_type="application/json",
        )
        set_client(
            RESTClientObject(
                auth=("u-**********", "p-******"), endpoint="https://host_name.com/api/v2"
            )
        )
        restored_client = get_client()
        assert restored_client.domain == "https://host_name.com"

    @pytest.mark.usefixtures("no_client_version_check")
    @mock.patch("datarobot.client._file_exists", return_value=False)
    def test_instantiation_from_env(self, mock_file_exists):
        """
        Test instantiation with creds from virtual environment
        """
        with patch.dict(
            "os.environ",
            {"DATAROBOT_API_TOKEN": "venv_token", "DATAROBOT_ENDPOINT": "https://host_name.com"},
        ):
            rest_client = get_client()
            assert rest_client.token == "venv_token"
            assert rest_client.endpoint == "https://host_name.com"

        set_client(None)

        with patch.dict(
            "os.environ",
            {
                "DATAROBOT_API_TOKEN": "venv_token",
                "DATAROBOT_ENDPOINT": "https://host_name.com",
                "DATAROBOT_MAX_RETRIES": "2",
            },
        ):
            rest_client = get_client()
            assert rest_client.token == "venv_token"
            assert rest_client.connect_timeout == DEFAULT_TIMEOUT.CONNECT
            assert rest_client.adapters
            for adapter in rest_client.adapters.values():
                assert adapter.max_retries.total == 2

    @pytest.mark.usefixtures("no_client_version_check")
    @mock.patch("datarobot.client._file_exists", return_value=False)
    def test_instantiation_from_env_checks_client_compatibility(
        self, mock_file_exists, no_client_version_check
    ):
        """
        Test instantiation with creds from virtual environment checks client compatibility.
        """
        with patch.dict(
            "os.environ",
            {"DATAROBOT_API_TOKEN": "venv_token", "DATAROBOT_ENDPOINT": "https://host_name.com"},
        ):
            rest_client = Client()
            assert rest_client.token == "venv_token"
            assert rest_client.endpoint == "https://host_name.com"
        assert no_client_version_check.called

    @pytest.mark.usefixtures("no_client_version_check")
    def test_instantiation_from_file_with_wrong_path(self):
        with patch.dict("os.environ", {"DATAROBOT_CONFIG_FILE": "./tests/fixtures/.datarobotrc"}):
            with pytest.raises(ValueError):
                get_client()

    @pytest.mark.usefixtures("no_client_version_check")
    def test_instantiation_from_yaml_file_api_token(self):
        file_data = "token: fake_token\nendpoint: https://host_name.com\nmax_retries: 3"
        config_file = tempfile.NamedTemporaryFile(delete=False, suffix=".yaml")
        try:
            with open(config_file.name, mode="w") as config:
                config.write(file_data)
            fake_environ = {"DATAROBOT_CONFIG_FILE": config_file.name}
            with patch("os.environ", fake_environ):
                rest_client = get_client()
            assert rest_client.token == "fake_token"
            assert rest_client.endpoint == "https://host_name.com"
            assert rest_client.connect_timeout == DEFAULT_TIMEOUT.CONNECT
            assert rest_client.adapters
            for adapter in rest_client.adapters.values():
                assert adapter.max_retries.total == 3
        finally:
            os.remove(config_file.name)

    @pytest.mark.usefixtures("no_client_version_check")
    def test_instantiation_from_yaml_file_set_timeout(self, temporary_file):
        connect_timeout = 55
        file_data = (
            "token: fake_token\nendpoint: https://host_name.com\nconnect_timeout: {}"
        ).format(connect_timeout)
        with open(temporary_file, mode="w") as config:
            config.write(file_data)
        fake_environ = {"DATAROBOT_CONFIG_FILE": temporary_file}
        with patch("os.environ", fake_environ):
            rest_client = get_client()
        assert rest_client.connect_timeout == connect_timeout

    @pytest.mark.usefixtures("no_client_version_check")
    def test_instantiation_from_yaml_checks_client_compatibility(
        self, temporary_file, no_client_version_check
    ):
        connect_timeout = 55
        file_data = (
            "token: fake_token\nendpoint: https://host_name.com\nconnect_timeout: {}"
        ).format(connect_timeout)
        with open(temporary_file, mode="w") as config:
            config.write(file_data)
        fake_environ = {"DATAROBOT_CONFIG_FILE": temporary_file}
        with patch("os.environ", fake_environ):
            Client()
        assert no_client_version_check.called

    @pytest.mark.usefixtures("no_client_version_check")
    def test_client_from_codeline(self):
        Client(token="some_token", endpoint="https://endpoint.com")
        c = get_client()
        assert c.token == "some_token"
        assert c.endpoint == "https://endpoint.com"
        assert c.connect_timeout == DEFAULT_TIMEOUT.CONNECT
        assert c.adapters
        for adapter in c.adapters.values():
            assert adapter.max_retries.total > 0
            assert adapter.max_retries.method_whitelist == Retry.DEFAULT_METHOD_WHITELIST
            assert adapter.max_retries.respect_retry_after_header

    @pytest.mark.usefixtures("no_client_version_check")
    def test_client_from_codeline_set_timeout(self):
        connect_timeout = 55
        Client(token="some_token", endpoint="https://endpoint.com", connect_timeout=connect_timeout)
        c = get_client()
        assert c.connect_timeout == connect_timeout

    @pytest.mark.usefixtures("no_client_version_check")
    def test_client_from_codeline_max_retries(self):
        Client(
            token="some token",
            endpoint="https://endpoint.com",
            max_retries=Retry(read=3, connect=3),
        )
        c = get_client()
        assert c.adapters
        for adapter in c.adapters.values():
            assert adapter.max_retries.read == 3
            assert adapter.max_retries.connect == 3

    @pytest.mark.usefixtures("no_client_version_check")
    def test_client_user_agent(self):
        with patch("platform.system") as mock_system, patch(
            "platform.release"
        ) as mock_release, patch("platform.machine") as mock_machine, patch(
            "platform.python_version"
        ) as mock_py_version:
            mock_system.return_value = "Prince"
            mock_release.return_value = "Party"
            mock_machine.return_value = "x19_99"
            mock_py_version.return_value = "3.14.159"
            header = RESTClientObject._make_user_agent_header()
        expected_agent = "DataRobotPythonClient/{} (Prince Party x19_99) Python-3.14.159".format(
            dr.__version__
        )
        assert header == {"User-Agent": expected_agent}

    @pytest.mark.usefixtures("no_client_version_check")
    def test_client_user_agent_with_suffix(self):
        suffix = "Blamo/8.23 (6.30)"
        orig = RESTClientObject._make_user_agent_header()
        with_suffix = RESTClientObject._make_user_agent_header(suffix=suffix)
        assert with_suffix["User-Agent"].startswith(orig["User-Agent"])
        assert with_suffix["User-Agent"].endswith(" " + suffix)


class RestErrors(SDKTestcase):
    @responses.activate
    def test_404_plain_text(self):
        """
        Bad request in plain text
        """
        raw = "Not Found"

        responses.add(
            responses.GET,
            "https://host_name.com/projects/404",
            body=raw,
            status=404,
            content_type="text/plain",
        )

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get("projects/404")

        self.assertEqual(str(app_error.exception), "404 client error: Not Found")

    @responses.activate
    def test_404_json(self):
        """
        Bad request with datarobot reason in json
        """
        raw = """
        {"message": "Not Found"}
        """

        responses.add(
            responses.GET,
            "https://host_name.com/projects/404",
            body=raw,
            status=404,
            content_type="application/json",
        )

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get("projects/404")

        exception_message = str(app_error.exception)
        self.assertIn("404 client error", exception_message)
        self.assertIn("Not Found", exception_message)

    @responses.activate
    def test_500_json(self):
        """
        Bad request with datarobot reason in json
        """
        raw = """
        {"message": "Not Found"}
        """

        responses.add(
            responses.GET,
            "https://host_name.com/projects/500",
            body=raw,
            status=500,
            content_type="application/json",
        )

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get("projects/500")

        exception_message = str(app_error.exception)
        self.assertIn("500 server error", exception_message)
        self.assertIn("Not Found", exception_message)

    @responses.activate
    def test_other_errors(self):
        """
        Other errors
        """
        raw = """
        {"message": "Bad response"}
        """

        responses.add(
            responses.GET,
            "https://host_name.com/projects/500",
            body=raw,
            status=500,
            content_type="application/json",
        )

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get("projects/500")
            self.assertEqual(
                str(app_error.exception), "Connection refused: https://host_name.com/projects/500",
            )


class TestClientAttributes(unittest.TestCase):
    def test_main_useful_things_under_datarobot(self):
        """To lower the intimidation factor, let's try to limit the objects
        that show up at the root of datarobot

        This way, when they are in IPython and do tab-completion they get a
        sense for what is available to tinker with at the top level
        """
        known_names = {
            "Project",
            "Model",
            "DatetimeModel",
            "PrimeModel",
            "BlenderModel",
            "FrozenModel",
            "Ruleset",
            "Blueprint",
            "BlueprintTaskDocument",
            "BlueprintChart",
            "ModelBlueprintChart",
            "ModelJob",
            "PredictJob",
            "Job",
            "Predictions",
            "PredictionDataset",
            "PrimeFile",
            "QUEUE_STATUS",
            "Client",
            "AUTOPILOT_MODE",
            "AppPlatformError",
            "utils",
            "errors",
            "models",
            "client",
            "rest",
            "SCORING_TYPE",
            "DatasetFeaturelist",
            "Featurelist",
            "ModelingFeaturelist",
            "Feature",
            "FeatureHistogram",
            "FeatureSettings",
            "ModelingFeature",
            "FeatureLineage",
            "helpers",
            "RandomCV",
            "StratifiedCV",
            "GroupCV",
            "UserCV",
            "RandomTVH",
            "UserTVH",
            "StratifiedTVH",
            "GroupTVH",
            "DatetimePartitioning",
            "DatetimePartitioningSpecification",
            "BacktestSpecification",
            "partitioning_methods",
            "AdvancedOptions",
            "Periodicity",
            "VERBOSITY_LEVEL",
            "enums",
            "ImportedModel",
            "ReasonCodesInitialization",
            "ReasonCodes",
            "PayoffMatrix",
            "PredictionExplanationsInitialization",
            "PredictionExplanations",
            "RatingTable",
            "RatingTableModel",
            "TARGET_TYPE",
            "TrainingPredictions",
            "TrainingPredictionsJob",
            "ShapImpact",
            "ShapMatrix",
            "ShapMatrixJob",
            "SharingAccess",
            "ModelRecommendation",
            "DataDriver",
            "Connector",
            "DataStore",
            "Dataset",
            "DatasetDetails",
            "DatasetFeature",
            "DatasetFeatureHistogram",
            "DataSource",
            "DataSourceParameters",
            "ComplianceDocumentation",
            "ComplianceDocTemplate",
            "CalendarFile",
            "PredictionServer",
            "Deployment",
            "SecondaryDatasetConfigurations",
            "BatchPredictionJob",
            "BatchPredictionJobDefinition",
            "Credential",
            "InteractionFeature",
            "ExternalScores",
            "ExternalLiftChart",
            "ExternalResidualsChart",
            "ExternalRocCurve",
            "ExternalMulticlassLiftChart",
            "ExternalConfusionChart",
            "ExecutionEnvironment",
            "ExecutionEnvironmentVersion",
            "CustomInferenceImage",
            "CustomInferenceModel",
            "CustomModelVersion",
            "CustomModelVersionDependencyBuild",
            "CustomModelTest",
            "FeatureImpactJob",
            "RelationshipsConfiguration",
            "NETWORK_EGRESS_POLICY",
            "FeatureAssociationMatrix",
            "FeatureAssociationMatrixDetails",
            "FeatureAssociationFeaturelists",
            "AutomatedDocument",
            "DatasetDefinition",
            "feature_discovery",
            "SecondaryDataset",
            "Relationship",
            "SNAPSHOT_POLICY",
        }

        found_names = {name for name in dir(dr) if not name.startswith("_")}
        assert found_names == known_names


@pytest.yield_fixture
def mock_os():
    with patch("datarobot.client.os", autospec=True) as m:
        yield m


@pytest.yield_fixture
def mock_file_exists():
    with patch("datarobot.client._file_exists") as m:
        yield m


@pytest.yield_fixture
def mock_os_environ():
    with patch("datarobot.client.os.environ") as m:
        yield m



import json
from uuid import uuid4

import pytest
import responses

from datarobot import PredictionServer


@pytest.fixture()
def prediction_server_data_cloud():
    return {
        "id": "5c2aad800000000000000000",
        "url": "https://predictions.example.com",
        "datarobot-key": str(uuid4()),
    }


@pytest.fixture()
def prediction_server_data_on_prem():
    return {
        "id": None,
        "url": "https://predictions.example.com",
        "datarobot-key": None,
    }


def make_prediction_server_list_response(prediction_server_data):
    return {"count": 1, "next": None, "previous": None, "data": [prediction_server_data]}


@responses.activate
@pytest.mark.parametrize(
    "prediction_server_data_fixture",
    ["prediction_server_data_cloud", "prediction_server_data_on_prem"],
)
def test_prediction_server_list(request, prediction_server_data_fixture):
    prediction_server_data = request.getfixturevalue(prediction_server_data_fixture)

    server_list_response = make_prediction_server_list_response(prediction_server_data)
    responses.add(
        responses.GET,
        "https://host_name.com/predictionServers/",
        status=200,
        content_type="application/json",
        body=json.dumps(server_list_response),
    )

    prediction_servers = PredictionServer.list()
    assert len(prediction_servers) == 1

    prediction_server = prediction_servers[0]
    assert prediction_server.id == prediction_server_data["id"]
    assert prediction_server.url == prediction_server_data["url"]
    assert prediction_server.datarobot_key == prediction_server_data["datarobot-key"]

    repr = "PredictionServer({url})".format(url=prediction_server_data["url"])
    assert str(prediction_server) == repr
